def interface____differentiated__3() -> None:
    return None

mod = "django.contrib.contenttypes.models"
summary = "__virtual__.mod_3961720227::django.contrib.contenttypes.models::installed_apps=__installed_apps_hash__::significant=2877817555"

import django.contrib.contenttypes.models
import django.db.models
ConcreteQuerySet__ContentType = django.db.models.QuerySet[django.contrib.contenttypes.models.ContentType]
Concrete__ContentType = django.contrib.contenttypes.models.ContentType



from django.db import models

from extended_mypy_django_plugin.django_analysis import (
    Field,
    ImportPath,
    Model,
    Project,
    discovery,
    protocols,
)


class TestConcreteModelsDiscovery:
    def test_it_collects_concrete_models(
        self, loaded_django_example: protocols.Loaded[Project]
    ) -> None:
        import djangoexample.exampleapp.models as mod
        import djangoexample.exampleapp2.models as mod2

        made: dict[type[models.Model], protocols.Model] = {}

        def create_model(model: type[models.Model]) -> protocols.Model:
            created = Model.create(field_creator=Field.create, model=model)
            made[model] = created
            return created

        all_models = {
            ImportPath.from_cls(model): create_model(model)
            for model in [
                mod.Parent,
                mod.Parent2,
                mod.Child1,
                mod.Child2,
                mod.Child3,
                mod.Child4,
                mod2.ChildOther,
                mod2.ChildOther2,
            ]
        }
        concrete_models = discovery.ConcreteModelsDiscovery[Project]()(
            loaded_django_example, all_models
        )

        assert concrete_models == {
            ImportPath.from_cls(mod.Parent): [
                made[mod.Child1],
                made[mod.Child2],
                made[mod.Child3],
                made[mod.Child4],
                made[mod2.ChildOther],
                made[mod2.ChildOther2],
            ],
            ImportPath.from_cls(mod.Parent2): [
                made[mod.Child3],
                made[mod.Child4],
            ],
            ImportPath.from_cls(mod.Child1): [made[mod.Child1]],
            ImportPath.from_cls(mod.Child2): [made[mod.Child2]],
            ImportPath.from_cls(mod.Child3): [made[mod.Child3]],
            ImportPath.from_cls(mod.Child4): [made[mod.Child4]],
            ImportPath.from_cls(mod2.ChildOther): [made[mod2.ChildOther]],
            ImportPath.from_cls(mod2.ChildOther2): [made[mod2.ChildOther2]],
        }



from ._plugin import protocols
from ._plugin.config import ExtraOptions
from ._plugin.entry import PluginProvider
from ._plugin.hook import Hook, hook
from ._plugin.plugin import ExtendedMypyStubs
from ._plugin.virtual_dependencies import VirtualDependencyHandler, VirtualDependencyHandlerBase

__all__ = [
    "hook",
    "Hook",
    "protocols",
    "ExtraOptions",
    "PluginProvider",
    "ExtendedMypyStubs",
    "VirtualDependencyHandler",
    "VirtualDependencyHandlerBase",
]



from __future__ import annotations

import dataclasses
import functools
import importlib
import operator
import os
import pathlib
import re
import shutil
import textwrap
from collections.abc import Callable, Iterator, Mapping, MutableMapping, Sequence, Set
from typing import TYPE_CHECKING, Generic, Literal, Protocol, TypeVar, cast

from .. import protocols
from ..discovery import ImportPath
from . import dependency

T_Report = TypeVar("T_Report", bound="Report")

regexes = {
    "mod_decl": re.compile(r'^mod = "(?P<mod>[^"]+)"$'),
    "summary_decl": re.compile(r'^summary = "(?P<summary>[^"]+)"$'),
}


@dataclasses.dataclass(frozen=True, kw_only=True)
class CombinedReport(Generic[protocols.T_Report]):
    version: str
    report: protocols.T_Report
    write_empty_virtual_dep: protocols.EmptyVirtualDepWriter

    def ensure_virtual_dependency(self, *, module_import_path: str) -> None:
        if module_import_path.startswith("django."):
            # Don't create empty virtual deps for django dependencies
            return

        # This is a heuristic that should be accurate enough to catch modules that contain models
        # Though it may miss some models and it may include modules that aren't related to django models
        if ".models." not in module_import_path and not module_import_path.endswith(".models"):
            return

        # An empty virtual dep is only written if there is no virtual dep to begin with
        virtual_import_path = self.write_empty_virtual_dep(
            module_import_path=protocols.ImportPath(module_import_path)
        )
        if virtual_import_path:
            self.report.register_module(
                module_import_path=protocols.ImportPath(module_import_path),
                virtual_import_path=virtual_import_path,
            )


@dataclasses.dataclass(frozen=True, kw_only=True)
class Report:
    concrete_annotations: MutableMapping[protocols.ImportPath, protocols.ImportPath] = (
        dataclasses.field(default_factory=dict)
    )
    concrete_querysets: MutableMapping[protocols.ImportPath, protocols.ImportPath] = (
        dataclasses.field(default_factory=dict)
    )
    report_import_path: MutableMapping[protocols.ImportPath, protocols.ImportPath] = (
        dataclasses.field(default_factory=dict)
    )

    def register_module(
        self,
        *,
        module_import_path: protocols.ImportPath,
        virtual_import_path: protocols.ImportPath,
    ) -> None:
        self.report_import_path[module_import_path] = virtual_import_path

    def register_model(
        self,
        *,
        model_import_path: protocols.ImportPath,
        virtual_import_path: protocols.ImportPath,
        concrete_name: str,
        concrete_queryset_name: str,
        concrete_models: Sequence[protocols.Model],
    ) -> None:
        module_import_path, name = ImportPath.split(model_import_path)

        self.concrete_annotations[model_import_path] = ImportPath(
            f"{virtual_import_path}.{concrete_name}"
        )
        self.concrete_querysets[model_import_path] = ImportPath(
            f"{virtual_import_path}.{concrete_queryset_name}"
        )

    def get_concrete_aliases(self, *models: str) -> Mapping[str, str | None]:
        result: dict[str, str | None] = {}
        for model in sorted(models):
            result[model] = self.concrete_annotations.get(protocols.ImportPath(model))
        return result

    def get_queryset_aliases(self, *models: str) -> Mapping[str, str | None]:
        result: dict[str, str | None] = {}
        for model in sorted(models):
            result[model] = self.concrete_querysets.get(protocols.ImportPath(model))
        return result

    def additional_deps(
        self,
        *,
        file_import_path: str,
        imports: Set[str],
        super_deps: Sequence[tuple[int, str, int]],
        django_settings_module: str,
        using_incremental_cache: bool,
    ) -> Sequence[tuple[int, str, int]]:
        if file_import_path.startswith("django."):
            # Don't add additional deps to django itself
            return super_deps

        report_names = set(self.report_import_path.values())
        if file_import_path in report_names:
            # Don't add additional deps to our virtual imports
            # if things they depend on change, then the virtual dep also changes already
            return super_deps

        # We need to include the virtual dependency so that we can find it and use
        # The type aliases it provides to resolve concrete annotations
        report_name = self.report_import_path.get(protocols.ImportPath(file_import_path))
        if report_name:
            extra_dep = (25, report_name, -1)
            if extra_dep not in super_deps:
                super_deps = [*super_deps, extra_dep]

        # When we're using the incremental cache we also want to make sure that
        # mypy understands there is a relationship between the file and the settings module
        # This isn't necessary in daemon mode cause changes to that will make us restart dmypy
        # And when there is no cache everything is from scratch anyways
        if report_name and using_incremental_cache:
            settings_dep = (25, django_settings_module, -1)
            if settings_dep not in super_deps:
                super_deps = [*super_deps, settings_dep]

        return super_deps


@dataclasses.dataclass(frozen=True, kw_only=True)
class RenderedVirtualDependency(Generic[protocols.T_Report]):
    content: str
    summary_hash: str | None
    report: protocols.T_Report
    virtual_import_path: protocols.ImportPath


@dataclasses.dataclass(frozen=True, kw_only=True)
class VirtualDependencyScribe(Generic[protocols.T_VirtualDependency, protocols.T_Report]):
    hasher: protocols.Hasher
    report_maker: protocols.ReportMaker[protocols.T_Report]
    virtual_dependency: protocols.T_VirtualDependency
    all_virtual_dependencies: protocols.VirtualDependencyMap[protocols.T_VirtualDependency]
    make_differentiator: Callable[[], str]
    installed_apps_hash: str

    @classmethod
    def make_empty_virtual_dependency_content(
        cls, *, module_import_path: protocols.ImportPath
    ) -> str:
        """
        An empty virtual dependency is literally a placeholder and doesn't contain information
        until the module it represents is an installed module.

        So we only need to contain a "mod" and "summary" variable that our "get_report_summary" method can find
        """
        return (
            textwrap.dedent(f"""
        mod = "{module_import_path}"
        summary = "||not_installed||"
        """).strip()
            + "\n"
        )

    def render(self) -> RenderedVirtualDependency[protocols.T_Report]:
        report = self.report_maker()
        summary_hash = self._get_summary_hash()

        module_import_path = self.virtual_dependency.summary.module_import_path
        virtual_import_path = self.virtual_dependency.summary.virtual_import_path
        report.register_module(
            module_import_path=module_import_path, virtual_import_path=virtual_import_path
        )

        content = self._template_virtual_dependency(
            report=report, virtual_import_path=virtual_import_path, summary_hash=summary_hash
        )

        return RenderedVirtualDependency(
            content=content,
            summary_hash=summary_hash,
            report=report,
            virtual_import_path=virtual_import_path,
        )

    @classmethod
    def get_report_summary(cls, location: pathlib.Path) -> str | None:
        """
        Given some location return the summary from that location.

        If the report doesn't have a summary or is for a module that doesn't exist anymore then return None
        """
        if not location.is_file():
            return None

        if location.suffix != ".py":
            return None

        # Look for 'mod = "{mod}"' and 'summary = "{summary}"' lines
        mod: str | None = None
        summary: str | None = None
        for line in location.read_text().splitlines():
            m = regexes["mod_decl"].match(line)
            if m:
                mod = m.groupdict()["mod"]

            m = regexes["summary_decl"].match(line)
            if m:
                summary = m.groupdict()["summary"]

            if mod and summary:
                break

        if mod is None or summary is None:
            # either no mod or not summary, so dependency is corrupt or irrelevant
            return None

        try:
            importlib.util.find_spec(mod)
        except ModuleNotFoundError:
            # If we can't import the module this represents, we assume it doesn't exist
            return None
        else:
            return summary

    def _get_summary_hash(self) -> str:
        summary = self.virtual_dependency.summary

        significant = self.hasher(*(info.encode() for info in summary.significant_info))

        return "::".join(
            [
                f"{summary.virtual_import_path}",
                str(summary.module_import_path),
                f"installed_apps={self.installed_apps_hash}",
                f"significant={significant}",
            ]
        )

    def _template_virtual_dependency(
        self,
        *,
        report: protocols.T_Report,
        virtual_import_path: protocols.ImportPath,
        summary_hash: str | None,
    ) -> str:
        module_import_path = self.virtual_dependency.summary.module_import_path
        summary = "None" if summary_hash is None else f'"{summary_hash}"'

        # mypy only considers a dependency as changed if it's public interface changes
        # Which is where either the static name or types change
        # So we include a function that has a different name everytime we write to the file
        # We rely on the contents of "summary" to not overwrite the same content with but
        # with a different interface when installing the dependencies
        content = textwrap.dedent(f"""
        def interface__{self.make_differentiator()}() -> None:
            return None

        mod = "{module_import_path}"
        summary = {summary}

        """)

        added_imports: set[protocols.ImportPath] = set()
        annotations: set[str] = set()

        for model, concrete in self.virtual_dependency.concrete_models.items():
            querysets: list[str] = []

            added_imports.add(model)
            for conc in sorted(concrete, key=operator.attrgetter("import_path")):
                added_imports.add(conc.import_path)
                if conc.default_custom_queryset:
                    added_imports.add(conc.default_custom_queryset)
                    querysets.append(conc.default_custom_queryset)
                else:
                    added_imports.add(ImportPath("django.db.models.QuerySet"))
                    querysets.append(f"django.db.models.QuerySet[{conc.import_path}]")

            ns, name = ImportPath.split(model)
            concrete_name = f"Concrete__{name}"
            queryset_name = f"ConcreteQuerySet__{name}"

            if concrete:
                annotations.add(
                    f"{concrete_name} = {' | '.join(conc.import_path for conc in concrete)}"
                )
            if querysets:
                annotations.add(f"{queryset_name} = {' | '.join(querysets)}")

            report.register_model(
                model_import_path=model,
                virtual_import_path=virtual_import_path,
                concrete_queryset_name=queryset_name,
                concrete_name=concrete_name,
                concrete_models=concrete,
            )

        sorted_added_imported_modules = sorted(
            {".".join(imp.split(".")[:-1]) for imp in added_imports}
        )

        # We add type aliases we use to resolve our concrete annotations to the dependency
        # This means that the mypy plugin relies completely on Django introspection to know
        # how to resolve the annotations, and we avoid problems around mypy not knowing about
        # relevant files when it analyses each file
        extra_lines = [
            *(f"import {import_path}" for import_path in sorted_added_imported_modules),
            *(f"{line}" for line in sorted(annotations)),
        ]

        if extra_lines:
            content = content + "\n".join(extra_lines)

        return content.strip() + "\n"


@dataclasses.dataclass(frozen=True, kw_only=True)
class ReportCombiner(Generic[T_Report]):
    reports: Sequence[T_Report]
    report_maker: protocols.ReportMaker[T_Report]

    def combine(
        self, *, version: str, write_empty_virtual_dep: protocols.EmptyVirtualDepWriter
    ) -> protocols.CombinedReport[T_Report]:
        final = self.report_maker()
        for report in self.reports:
            final.concrete_annotations.update(report.concrete_annotations)
            final.concrete_querysets.update(report.concrete_querysets)
            final.report_import_path.update(report.report_import_path)

        return CombinedReport(
            version=version, report=final, write_empty_virtual_dep=write_empty_virtual_dep
        )


class ReportSummaryGetter(Protocol):
    """
    Protocol for a callable that returns a summary from a path

    Where None is returned if the path isn't a valid virtual dependency

    And a string is the summary from that virtual dependency
    """

    def __call__(self, location: pathlib.Path, /) -> str | None: ...


@dataclasses.dataclass(frozen=True, kw_only=True)
class ReportInstaller:
    _written: dict[pathlib.Path, str | Literal[False] | None] = dataclasses.field(
        init=False, default_factory=dict
    )
    _get_report_summary: ReportSummaryGetter

    def write_report(
        self,
        *,
        scratch_root: pathlib.Path,
        summary_hash: str | Literal[False] | None,
        virtual_import_path: protocols.ImportPath,
        content: str,
    ) -> bool:
        location = scratch_root / f"{virtual_import_path.replace('.', os.sep)}.py"
        if not location.is_relative_to(scratch_root):
            raise RuntimeError(
                f"Virtual dependency ends up being outside of the scratch root: {virtual_import_path}"
            )

        if location.exists() and summary_hash is False:
            return False

        location.parent.mkdir(parents=True, exist_ok=True)
        location.write_text(content)
        self._written[location] = summary_hash
        return True

    def install_reports(
        self,
        *,
        scratch_root: pathlib.Path,
        destination: pathlib.Path,
        virtual_namespace: protocols.ImportPath,
    ) -> None:
        virtual_destination = destination / virtual_namespace
        virtual_destination.mkdir(parents=True, exist_ok=True)

        seen: set[pathlib.Path] = set()

        # For all the dependencies we have written to the filesystem, determine
        # if they represent different information to what is already on the destination
        # and move across those that are different
        for location, summary in self._written.items():
            relative_path = location.relative_to(scratch_root)
            destination_path = destination / relative_path

            seen.add(destination_path)
            found_summary: str | None = None
            if destination_path.exists():
                found_summary = self._get_report_summary(destination_path)

            if found_summary != summary:
                destination_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.move(location, destination_path)

        # Then we go ahead and do some garbage collection on the destination
        # So that the destination is only ever dependencies for modules that exist
        # and we don't have an infinitely growing folder of virtual dependencies
        for root, dirs, files in os.walk(virtual_destination):
            for name in list(dirs):
                location = pathlib.Path(root) / name
                if location not in seen:
                    if self._get_report_summary(location) is None:
                        shutil.rmtree(location)
                        dirs.remove(name)

            for name in files:
                location = pathlib.Path(root) / name
                if location not in seen:
                    if self._get_report_summary(location) is None:
                        location.unlink(missing_ok=True)


@dataclasses.dataclass(frozen=True, kw_only=True)
class ReportFactory(Generic[protocols.T_VirtualDependency, protocols.T_Report]):
    hasher: protocols.Hasher

    report_installer: protocols.ReportInstaller
    report_combiner_maker: protocols.ReportCombinerMaker[protocols.T_Report]
    report_maker: protocols.ReportMaker[protocols.T_Report]
    make_empty_virtual_dependency_content: protocols.MakeEmptyVirtualDepContent
    report_scribe: protocols.VirtualDependencyScribe[
        protocols.T_VirtualDependency, protocols.T_Report
    ]

    def deploy_scribes(
        self, virtual_dependencies: protocols.VirtualDependencyMap[protocols.T_VirtualDependency]
    ) -> Iterator[protocols.RenderedVirtualDependency[protocols.T_Report]]:
        for virtual_dependency in virtual_dependencies.values():
            yield self.report_scribe(
                virtual_dependency=virtual_dependency,
                all_virtual_dependencies=virtual_dependencies,
            )

    def determine_version(
        self,
        *,
        destination: pathlib.Path,
        virtual_namespace: protocols.ImportPath,
        project_version: str,
        written_dependencies: Sequence[protocols.RenderedVirtualDependency[protocols.T_Report]],
    ) -> str:
        virtual_dep_hash = self.hasher(
            *(
                f"{written.virtual_import_path}:{written.summary_hash}".encode()
                for written in written_dependencies
            )
        )
        return f"{virtual_namespace}|{project_version}|written_deps:{virtual_dep_hash}"


def make_report_factory(
    *,
    hasher: protocols.Hasher,
    report_maker: protocols.ReportMaker[Report],
    installed_apps_hash: str,
    make_differentiator: Callable[[], str],
) -> protocols.ReportFactory[protocols.T_VirtualDependency, Report]:
    """
    Make a ReportFactory that's specific to the our implementation of protocols.Report found here
    """

    def report_scribe(
        *,
        virtual_dependency: protocols.T_VirtualDependency,
        all_virtual_dependencies: protocols.VirtualDependencyMap[protocols.T_VirtualDependency],
    ) -> protocols.RenderedVirtualDependency[Report]:
        return VirtualDependencyScribe(
            hasher=hasher,
            report_maker=Report,
            installed_apps_hash=installed_apps_hash,
            virtual_dependency=virtual_dependency,
            all_virtual_dependencies=all_virtual_dependencies,
            make_differentiator=make_differentiator,
        ).render()

    return ReportFactory(
        hasher=hasher,
        report_maker=report_maker,
        report_scribe=report_scribe,
        report_installer=ReportInstaller(
            _get_report_summary=VirtualDependencyScribe.get_report_summary
        ),
        report_combiner_maker=functools.partial(ReportCombiner, report_maker=report_maker),
        make_empty_virtual_dependency_content=VirtualDependencyScribe.make_empty_virtual_dependency_content,
    )


if TYPE_CHECKING:
    C_Report = Report
    C_CombinedReport = CombinedReport[Report]
    C_ReportFactory = ReportFactory[dependency.C_VirtualDependency, C_Report]
    C_ReportCombiner = ReportCombiner[C_Report]
    C_ReportInstaller = ReportInstaller
    C_RenderedVirtualDependency = RenderedVirtualDependency[C_Report]

    _R: protocols.P_Report = cast(Report, None)
    _RC: protocols.P_CombinedReport = cast(CombinedReport[protocols.P_Report], None)
    _RF: protocols.P_ReportFactory = cast(
        ReportFactory[protocols.P_VirtualDependency, protocols.P_Report], None
    )
    _WVD: protocols.P_RenderedVirtualDependency = cast(
        RenderedVirtualDependency[protocols.P_Report], None
    )
    _RI: protocols.P_ReportInstaller = cast(ReportInstaller, None)
    _MEVDC: protocols.MakeEmptyVirtualDepContent = (
        VirtualDependencyScribe.make_empty_virtual_dependency_content
    )

    _CRC: protocols.ReportCombiner[C_Report] = cast(C_ReportCombiner, None)
    _CRCC: protocols.CombinedReport[C_Report] = cast(C_CombinedReport, None)
    _CRF: protocols.ReportFactory[dependency.C_VirtualDependency, C_Report] = cast(
        C_ReportFactory, None
    )
    _CRM: protocols.ReportMaker[C_Report] = C_Report
    _CWVD: protocols.RenderedVirtualDependency[C_Report] = cast(C_RenderedVirtualDependency, None)



import abc
import dataclasses
import functools
import pathlib
from typing import TYPE_CHECKING, Generic

from ..django_analysis import Project, discovery, project, virtual_dependencies
from ..django_analysis import protocols as d_protocols
from . import protocols as p_protocols


@dataclasses.dataclass(frozen=True, kw_only=True)
class VirtualDependencyHandlerBase(
    Generic[d_protocols.T_Project],
    virtual_dependencies.VirtualDependencyHandler[
        d_protocols.T_Project,
        virtual_dependencies.VirtualDependency[d_protocols.T_Project],
        virtual_dependencies.Report,
    ],
    abc.ABC,
):
    def get_report_maker(self) -> d_protocols.ReportMaker[virtual_dependencies.Report]:
        return virtual_dependencies.Report

    def make_report_factory(
        self, *, installed_apps_hash: str
    ) -> d_protocols.ReportFactory[
        virtual_dependencies.VirtualDependency[d_protocols.T_Project], virtual_dependencies.Report
    ]:
        return virtual_dependencies.make_report_factory(
            hasher=self.hasher,
            report_maker=self.get_report_maker(),
            installed_apps_hash=installed_apps_hash,
            make_differentiator=self.interface_differentiator,
        )

    def virtual_dependency_maker(
        self, *, virtual_dependency_namer: d_protocols.VirtualDependencyNamer
    ) -> d_protocols.VirtualDependencyMaker[
        d_protocols.T_Project, virtual_dependencies.VirtualDependency[d_protocols.T_Project]
    ]:
        return functools.partial(
            virtual_dependencies.VirtualDependency.create,
            discovered_project=self.discovered,
            virtual_dependency_namer=virtual_dependency_namer,
        )


@dataclasses.dataclass(frozen=True, kw_only=True)
class VirtualDependencyHandler(VirtualDependencyHandlerBase[Project]):
    @classmethod
    def make_project(cls, *, project_root: pathlib.Path, django_settings_module: str) -> Project:
        return Project(
            root_dir=project_root,
            additional_sys_path=[str(project_root)],
            env_vars={"DJANGO_SETTINGS_MODULE": django_settings_module},
            discovery=discovery.Discovery(),
        )


if TYPE_CHECKING:
    P_VirtualDependencyHandler = VirtualDependencyHandlerBase[project.C_Project]
    C_VirtualDependencyHandler = VirtualDependencyHandler

    _DVDH: p_protocols.P_VirtualDependencyHandler = VirtualDependencyHandlerBase[
        d_protocols.P_Project
    ].create_report
    _CDVDH: p_protocols.VirtualDependencyHandler[virtual_dependencies.Report] = (
        C_VirtualDependencyHandler.create_report
    )



import pathlib
import sys
import textwrap

import pytest
import pytest_mypy_plugins.utils
from extended_mypy_django_plugin_test_driver import OutputBuilder, Scenario, assertions
from pytest_mypy_plugins import OutputChecker


class TestErrors:
    def test_it_complains_if_annotating_a_typevar(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after(installed_apps=["example"])
        def _(expected: OutputBuilder) -> None:
            scenario.file(expected, "example/__init__.py", "")

            scenario.file(
                expected,
                "example/apps.py",
                """
                from django.apps import AppConfig

                class Config(AppConfig):
                    name = "example"
                """,
            )

            scenario.file(
                expected,
                "example/models.py",
                """
                from __future__ import annotations

                from django.db import models
                from typing import TypeVar
                from typing_extensions import Self
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                T_Leader = TypeVar("T_Leader", bound="Concrete[Leader]")

                class Leader(models.Model):
                    @classmethod
                    def new(cls) -> Concrete[Self]:
                        # ^ ERROR(misc) ^ Using a concrete annotation on a TypeVar is not currently supported
                        raise NotImplementedError()

                    class Meta:
                        abstract = True

                class Follower1(Leader):
                    pass

                def make_leader(model: type[T_Leader]) -> Concrete[T_Leader]:
                    # ^ ERROR(misc) ^ Using a concrete annotation on a TypeVar is not currently supported
                    raise NotImplementedError()

                def make_qs(model: type[T_Leader]) -> DefaultQuerySet[T_Leader]:
                    # ^ ERROR(misc) ^ Using a concrete annotation on a TypeVar is not currently supported
                    raise NotImplementedError()
                """,
            )

    def test_gracefully_handles_determine_version_failure_on_startup(
        self, scenario: Scenario, tmp_path: pathlib.Path
    ) -> None:
        if not scenario.for_daemon:
            pytest.skip("Test only relevant for the daemon")

        plugin_provider = tmp_path / "plugin.py"

        plugin_provider.write_text(
            textwrap.dedent("""
            import pathlib

            from extended_mypy_django_plugin.django_analysis import Project
            from extended_mypy_django_plugin.plugin import PluginProvider, VirtualDependencyHandler, ExtendedMypyStubs


            class VirtualDependencyHandler(VirtualDependencyHandler):
                @classmethod
                def make_project(
                    cls, *, project_root: pathlib.Path, django_settings_module: str
                ) -> Project:
                    raise ValueError("Computer says no")


            plugin = PluginProvider(ExtendedMypyStubs, VirtualDependencyHandler.create_report, locals())
            """)
        )

        scenario.scenario.additional_mypy_config = textwrap.dedent(
            f"""
            [mypy]
            plugins = {plugin_provider}

            [mypy.plugins.django-stubs]
            django_settings_module = mysettings
            """
        )

        with pytest.raises(pytest_mypy_plugins.utils.TypecheckAssertionError) as err:

            @scenario.run_and_check_mypy_after
            def _(expected: OutputBuilder) -> None:
                pass

        assert err.value.mypy_output is not None

        assertions.assert_glob_lines(
            err.value.mypy_output,
            f"""
            Error constructing plugin instance of Plugin
            
            Daemon crashed!
            Traceback (most recent call last):
            File "*extended_mypy_django_plugin/_plugin/plugin.py", line *, in make_virtual_dependency_report
            File "{plugin_provider}", line *, in make_project
            ValueError: Computer says no
            """,
        )

    def test_gracefully_handles_determine_version_failure_on_subsequent_run(
        self, scenario: Scenario, tmp_path: pathlib.Path
    ) -> None:
        if not scenario.for_daemon:
            pytest.skip("Test only relevant for the daemon")

        plugin_provider = tmp_path / "plugin.py"
        marker = tmp_path / "marker"
        marker2 = tmp_path / "marker2"

        # pytest plugin I use needs work which is under way but in the meantime I must hack around
        # how inside the test I can't turn off the auto second try
        marker.write_text("")
        marker2.write_text("")

        # Changing the contents of this file will trigger the daemon to restart
        # So we instead rely on the existence or absence of a file to trigger the error
        plugin_provider.write_text(
            textwrap.dedent(f"""
            import pathlib

            from extended_mypy_django_plugin.django_analysis import Project
            from extended_mypy_django_plugin.main import PluginProvider, VirtualDependencyHandler, ExtendedMypyStubs


            class VirtualDependencyHandler(VirtualDependencyHandler):
                @classmethod
                def make_project(
                    cls, *, project_root: pathlib.Path, django_settings_module: str
                ) -> Project:
                    if pathlib.Path("{marker}").exists():
                        pathlib.Path("{marker}").unlink()
                        return super().make_project(
                            project_root=project_root,
                            django_settings_module=django_settings_module,
                        )

                    if pathlib.Path("{marker2}").exists():
                        pathlib.Path("{marker2}").unlink()
                        return super().make_project(
                            project_root=project_root,
                            django_settings_module=django_settings_module,
                        )

                    # Make this only fail on the startup to show if the run after restart works then
                    # then this failing doesn't break the daemon
                    pathlib.Path("{marker}").write_text('')
                    raise ValueError("Computer says no")


            plugin = PluginProvider(ExtendedMypyStubs, VirtualDependencyHandler.create_report, locals())
        """)
        )

        scenario.scenario.additional_mypy_config = textwrap.dedent(
            f"""
            [mypy]
            plugins = {plugin_provider}

            [mypy.plugins.django-stubs]
            django_settings_module = mysettings
            """
        )

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            pass

        called: list[int] = []

        class CheckNoCrashShowsFailure(OutputChecker):
            def check(self, ret_code: int, stdout: str, stderr: str) -> None:
                called.append(ret_code)

                assert ret_code == 0
                command = (
                    f"{sys.executable} -m extended_mypy_django_plugin.scripts.determine_django_state"
                    f" --config-file {scenario.scenario.execution_path}/mypy.ini"
                    f" --mypy-plugin {plugin_provider}"
                    " --version-file *"
                )

                assertions.assert_glob_lines(
                    stdout + stderr,
                    f"""
                    Failed to determine information about the django setup

                    > {command}
                    |
                    | Traceback (most recent call last):
                    |   File "{plugin_provider}", line *, in make_project
                    |     raise ValueError("Computer says no")
                    | ValueError: Computer says no
                    |
                    """,
                )

        scenario.run_and_check_mypy(scenario.expected, OutputCheckerKls=CheckNoCrashShowsFailure)
        assert called == [0]

        class CheckNoOutput(OutputChecker):
            def check(self, ret_code: int, stdout: str, stderr: str) -> None:
                called.append(ret_code)

                assert ret_code == 0
                assert stdout + stderr == ""

        marker.write_text("")
        scenario.run_and_check_mypy(scenario.expected, OutputCheckerKls=CheckNoOutput)
        assert called == [0, 0]

    def test_knowing_types_of_fields_on_parent_classes(self, scenario: Scenario) -> None:
        """
        This is a regression test to ensure that get_additional_deps doesn't cause class
        definitions to not understand parent types
        """

        @scenario.run_and_check_mypy_after(installed_apps=["example", "example2"])
        def _(expected: OutputBuilder) -> None:
            for app in ("example", "example2"):
                scenario.file(expected, f"{app}/__init__.py", "")
                scenario.file(
                    expected,
                    f"{app}/apps.py",
                    f"""
                    from django.apps import AppConfig

                    class Config(AppConfig):
                        name = "{app}"
                    """,
                )

            scenario.file(
                expected,
                "example/models/__init__.py",
                """
                from .parent import Parent 
                """,
            )
            scenario.file(
                expected,
                "example2/models/__init__.py",
                """
                from .children import Child
                """,
            )

            scenario.file(
                expected,
                "example/models/parent.py",
                """
                from django.db import models

                class Parent(models.Model):
                    response_body = models.TextField(max_length=12, blank=True)

                    class Meta:
                        abstract = True
                """,
            )

            scenario.file(
                expected,
                "example2/models/children.py",
                """
                from example import models as common_models
                from typing import TYPE_CHECKING
                from django.db import models

                class Child(common_models.Parent):
                    response_body = models.BooleanField()
                    # ^ ERROR(assignment) ^ Incompatible types in assignment (expression has type "BooleanField[bool | Combinable, bool]", base class "Parent" defined the type as "TextField[str | Combinable, str]")
                """,
            )



def interface____differentiated__6() -> None:
    return None

mod = "djangoexample.exampleapp2.models"
summary = "__virtual__.mod_3537308831::djangoexample.exampleapp2.models::installed_apps=__installed_apps_hash__::significant=1778215230"

import django.db.models
import djangoexample.exampleapp2.models
ConcreteQuerySet__ChildOther = django.db.models.QuerySet[djangoexample.exampleapp2.models.ChildOther]
ConcreteQuerySet__ChildOther2 = django.db.models.QuerySet[djangoexample.exampleapp2.models.ChildOther2]
Concrete__ChildOther = djangoexample.exampleapp2.models.ChildOther
Concrete__ChildOther2 = djangoexample.exampleapp2.models.ChildOther2



import abc
import dataclasses
import pathlib
import tempfile
import time
from typing import TYPE_CHECKING, Generic

from typing_extensions import Self

from ...version import VERSION
from .. import discovery, hasher, project, protocols
from . import dependency, report
from .folder import VirtualDependencyGenerator, VirtualDependencyInstaller
from .namer import VirtualDependencyNamer


@dataclasses.dataclass(frozen=True, kw_only=True)
class VirtualDependencyHandler(
    Generic[protocols.T_Project, protocols.T_VirtualDependency, protocols.T_Report], abc.ABC
):
    """
    This brings together the project, virtual dependencies and the report such that given a project
    we can generate the virtual dependencies and a relevant report.

    These three things are all customizable and this class is an orchestration mechanism.

    Usage is via the "create" and "create_report" classmethods, where "create_report" is a shortcut
    to saying ``Handler.create().make_report()``.

    And "create" is a shortcut to creating an instance of the Handler with the hasher and project
    found using the "make_hasher" and "make_project" classmethods on this class.
    """

    hasher: protocols.Hasher
    discovered: protocols.Discovered[protocols.T_Project]

    @classmethod
    def create(cls, *, project_root: pathlib.Path, django_settings_module: str) -> Self:
        return cls(
            hasher=cls.make_hasher(),
            discovered=(
                cls.make_project(
                    project_root=project_root, django_settings_module=django_settings_module
                )
                .load_project()
                .perform_discovery()
            ),
        )

    @classmethod
    def create_report(
        cls,
        *,
        project_root: pathlib.Path,
        django_settings_module: str,
        virtual_deps_destination: pathlib.Path,
    ) -> protocols.CombinedReport[protocols.T_Report]:
        return cls.create(
            project_root=project_root, django_settings_module=django_settings_module
        ).make_report(virtual_deps_destination=virtual_deps_destination)

    def make_report(
        self, virtual_deps_destination: pathlib.Path
    ) -> protocols.CombinedReport[protocols.T_Report]:
        """
        The main orchestration to create the virtual dependencies and the final
        combined report.

        There are a number of customization points that this class has hooks for so that
        this method remains generic to how virtual dependencies are represented on disk and
        in the report, as well as what information goes into them, and how and where they
        are written to disk.
        """
        installed_apps_hash = self.hash_installed_apps()
        settings_types_hash = self.hash_settings_types()
        virtual_namespace = self.get_virtual_namespace()
        virtual_dependency_namer = self.make_virtual_dependency_namer(
            virtual_namespace=virtual_namespace
        )
        virtual_dependency_maker = self.virtual_dependency_maker(
            virtual_dependency_namer=virtual_dependency_namer
        )
        all_virtual_dependencies = self.get_virtual_dependencies(
            virtual_dependency_maker=virtual_dependency_maker
        )
        report_factory = self.make_report_factory(installed_apps_hash=installed_apps_hash)
        project_version = f"plugin:{VERSION}:installed_apps:{installed_apps_hash}|settings_types:{settings_types_hash}"
        virtual_dependency_installer = self.make_virtual_dependency_installer(
            virtual_dependency_namer=virtual_dependency_namer,
            project_version=project_version,
            all_virtual_dependencies=all_virtual_dependencies,
        )

        with tempfile.TemporaryDirectory() as scratch_root:
            return virtual_dependency_installer(
                scratch_root=pathlib.Path(scratch_root),
                destination=virtual_deps_destination,
                virtual_namespace=virtual_namespace,
                report_factory=report_factory,
            )

    @classmethod
    @abc.abstractmethod
    def make_project(
        cls, *, project_root: pathlib.Path, django_settings_module: str
    ) -> protocols.T_Project: ...

    @abc.abstractmethod
    def make_report_factory(
        self, *, installed_apps_hash: str
    ) -> protocols.ReportFactory[protocols.T_VirtualDependency, protocols.T_Report]: ...

    @abc.abstractmethod
    def virtual_dependency_maker(
        self, *, virtual_dependency_namer: protocols.VirtualDependencyNamer
    ) -> protocols.VirtualDependencyMaker[protocols.T_Project, protocols.T_VirtualDependency]: ...

    def make_virtual_dependency_installer(
        self,
        *,
        project_version: str,
        virtual_dependency_namer: protocols.VirtualDependencyNamer,
        all_virtual_dependencies: protocols.VirtualDependencyMap[protocols.T_VirtualDependency],
    ) -> protocols.VirtualDependencyInstaller[protocols.T_VirtualDependency, protocols.T_Report]:
        return VirtualDependencyInstaller(
            virtual_dependency_namer=virtual_dependency_namer,
            project_version=project_version,
            virtual_dependencies=all_virtual_dependencies,
        )

    @classmethod
    def make_hasher(cls) -> protocols.Hasher:
        return hasher.adler32_hash

    def interface_differentiator(self) -> str:
        return str(time.time()).replace(".", "_")

    def hash_installed_apps(self) -> str:
        return self.hasher(
            *(app.encode() for app in self.discovered.loaded_project.settings.INSTALLED_APPS)
        )

    def hash_settings_types(self) -> str:
        return self.hasher(
            *(f"{k}:{v}".encode() for k, v in sorted(self.discovered.settings_types.items()))
        )

    def make_virtual_dependency_namer(
        self, *, virtual_namespace: protocols.ImportPath
    ) -> protocols.VirtualDependencyNamer:
        return VirtualDependencyNamer(namespace=virtual_namespace, hasher=self.hasher)

    def get_virtual_namespace(self) -> protocols.ImportPath:
        return discovery.ImportPath("__virtual_extended_mypy_django_plugin_report__")

    def get_virtual_dependencies(
        self,
        *,
        virtual_dependency_maker: protocols.VirtualDependencyMaker[
            protocols.T_Project, protocols.T_VirtualDependency
        ],
    ) -> protocols.VirtualDependencyMap[protocols.T_VirtualDependency]:
        return VirtualDependencyGenerator(virtual_dependency_maker=virtual_dependency_maker)(
            discovered_project=self.discovered
        )


if TYPE_CHECKING:
    C_VirtualDependencyHandler = VirtualDependencyHandler[
        project.C_Project, dependency.C_VirtualDependency, report.C_Report
    ]

    _VDH: protocols.P_VirtualDependencyHandler = VirtualDependencyHandler[
        protocols.P_Project, protocols.P_VirtualDependency, protocols.P_Report
    ].create_report

    _CVDH: protocols.VirtualDependencyHandler[report.C_Report] = (
        C_VirtualDependencyHandler.create_report
    )



from django.db import models

from djangoexample.relations1 import models as relations1


class Thing(models.Model):
    concrete = models.OneToOneField(
        relations1.Concrete1, related_name="thing", on_delete=models.CASCADE
    )



from mypy.nodes import CallExpr, MemberExpr
from mypy.plugin import (
    AttributeContext,
    FunctionContext,
    MethodContext,
)
from mypy.types import (
    AnyType,
    Instance,
    TypeOfAny,
    TypeType,
    TypeVarType,
    UnionType,
    get_proper_type,
)
from mypy.types import Type as MypyType

from . import protocols


class TypeChecking:
    def __init__(self, *, make_resolver: protocols.ResolverMaker) -> None:
        self.make_resolver = make_resolver

    def modify_cast_as_concrete(self, ctx: FunctionContext | MethodContext) -> MypyType:
        if len(ctx.arg_types) != 1:
            ctx.api.fail("Concrete.cast_as_concrete takes only one argument", ctx.context)
            return AnyType(TypeOfAny.from_error)

        if not ctx.arg_types[0]:
            ctx.api.fail("Mypy failed to tell us the type of the first argument", ctx.context)
            return AnyType(TypeOfAny.from_error)

        first_arg = get_proper_type(ctx.arg_types[0][0])
        if isinstance(first_arg, AnyType):
            ctx.api.fail("Failed to determine the type of the first argument", ctx.context)
            return AnyType(TypeOfAny.from_error)

        is_type: bool = False
        if isinstance(first_arg, TypeType):
            is_type = True
            first_arg = first_arg.item

        instances: list[Instance] = []
        if isinstance(first_arg, TypeVarType):
            if first_arg.values:
                for found in first_arg.values:
                    item = get_proper_type(found)
                    if isinstance(item, Instance):
                        instances.append(item)
                    else:
                        ctx.api.fail(
                            f"A value in the type var ({first_arg}) is unexpected: {item}: {type(item)}",
                            ctx.context,
                        )
                        return AnyType(TypeOfAny.from_error)
            else:
                item = get_proper_type(first_arg.upper_bound)
                if not isinstance(item, Instance):
                    ctx.api.fail(
                        f"Upper bound for type var ({first_arg}) is unexpected: {item}: {type(item)}",
                        ctx.context,
                    )
                    return AnyType(TypeOfAny.from_error)
                instances.append(item)

        elif isinstance(first_arg, Instance):
            instances.append(first_arg)

        elif isinstance(first_arg, UnionType):
            union_items = [get_proper_type(item) for item in first_arg.items]
            union_pairs = [
                (isinstance(part, TypeType), isinstance(part, Instance), part)
                for part in union_items
            ]
            are_all_instances = all(
                is_type or is_instance for is_type, is_instance, _ in union_pairs
            )
            if are_all_instances:
                for part in union_items:
                    found = part
                    if isinstance(found, TypeType):
                        is_type = True
                        found = found.item
                    if not isinstance(part, Instance):
                        are_all_instances = False
                        break
                    instances.append(part)

            if not are_all_instances:
                ctx.api.fail(
                    f"Expected only `type[MyClass]` or `MyClass` in a union provided to cast_as_concrete, got {union_items}",
                    ctx.context,
                )
                return AnyType(TypeOfAny.from_error)
        else:
            ctx.api.fail(
                f"cast_as_concrete must take a variable with a clear type, got {first_arg}: ({type(first_arg)})",
                ctx.context,
            )
            return AnyType(TypeOfAny.from_error)

        resolver = self.make_resolver(ctx=ctx)
        resolved = resolver.resolve(
            protocols.KnownAnnotations.CONCRETE, UnionType(tuple(instances))
        )
        if not resolved:
            # Error would have already been sent out
            return AnyType(TypeOfAny.from_error)

        if isinstance(resolved, UnionType):
            if is_type:
                resolved = UnionType(tuple(TypeType(item) for item in resolved.items))
        elif is_type:
            resolved = TypeType(resolved)

        return resolved

    def extended_get_attribute_resolve_manager_method(
        self,
        ctx: AttributeContext,
        *,
        resolve_manager_method_from_instance: protocols.ResolveManagerMethodFromInstance,
    ) -> MypyType:
        """
        Copied from django-stubs after https://github.com/typeddjango/django-stubs/pull/2027

        A 'get_attribute_hook' that is intended to be invoked whenever the TypeChecker encounters
        an attribute on a class that has 'django.db.models.BaseManager' as a base.
        """
        # Skip (method) type that is currently something other than Any of type `implementation_artifact`
        default_attr_type = get_proper_type(ctx.default_attr_type)
        if not isinstance(default_attr_type, AnyType):
            return default_attr_type
        elif default_attr_type.type_of_any != TypeOfAny.implementation_artifact:
            return default_attr_type

        # (Current state is:) We wouldn't end up here when looking up a method from a custom _manager_.
        # That's why we only attempt to lookup the method for either a dynamically added or reverse manager.
        if isinstance(ctx.context, MemberExpr):
            method_name = ctx.context.name
        elif isinstance(ctx.context, CallExpr) and isinstance(ctx.context.callee, MemberExpr):
            method_name = ctx.context.callee.name
        else:
            ctx.api.fail("Unable to resolve return type of queryset/manager method", ctx.context)
            return AnyType(TypeOfAny.from_error)

        if isinstance(ctx.type, Instance):
            return resolve_manager_method_from_instance(
                instance=ctx.type, method_name=method_name, ctx=ctx
            )
        elif isinstance(ctx.type, UnionType) and all(
            isinstance(get_proper_type(instance), Instance) for instance in ctx.type.items
        ):
            items: list[Instance] = []
            for instance in ctx.type.items:
                inst = get_proper_type(instance)
                if isinstance(inst, Instance):
                    items.append(inst)

            resolved = tuple(
                resolve_manager_method_from_instance(
                    instance=inst, method_name=method_name, ctx=ctx
                )
                for inst in items
            )
            return UnionType(resolved)
        else:
            ctx.api.fail(
                f'Unable to resolve return type of queryset/manager method "{method_name}"',
                ctx.context,
            )
            return AnyType(TypeOfAny.from_error)



def interface____differentiated__10() -> None:
    return None

mod = "djangoexample.empty_models.models"
summary = "__virtual__.mod_3808300370::djangoexample.empty_models.models::installed_apps=__installed_apps_hash__::significant=1232670738"



import dataclasses
import functools
import importlib
import inspect
import types
from collections import defaultdict
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Any, Generic, Protocol, cast

from django.db import models

from .. import protocols
from .import_path import ImportPath


class ModuleCreator(Protocol):
    def __call__(
        self,
        *,
        import_path: protocols.ImportPath,
        module: types.ModuleType | None,
        models: Sequence[type[models.Model]],
    ) -> protocols.Module: ...


def make_module_creator() -> ModuleCreator:
    from extended_mypy_django_plugin import django_analysis

    field_creator = django_analysis.Field.create
    model_creator = functools.partial(django_analysis.Model.create, field_creator=field_creator)
    return functools.partial(django_analysis.Module.create, model_creator=model_creator)


@dataclasses.dataclass(frozen=True, kw_only=True)
class DefaultInstalledModulesDiscovery(Generic[protocols.T_Project]):
    module_creator: ModuleCreator = dataclasses.field(default_factory=make_module_creator)

    def __call__(
        self, loaded_project: protocols.Loaded[protocols.T_Project], /
    ) -> protocols.ModelModulesMap:
        found: dict[protocols.ImportPath, list[type[models.Model]]] = defaultdict(list)
        for concrete_model_cls in loaded_project.apps.get_models():
            found[ImportPath.cls_module(concrete_model_cls)].append(concrete_model_cls)

        result: dict[protocols.ImportPath, protocols.Module] = {}

        for module_import_path, module_models in found.items():
            module = inspect.getmodule(module_models[0])
            if module is None:
                raise RuntimeError("Failed to determine the module these models appear in")

            entity = self.module_creator(
                import_path=module_import_path,
                module=module,
                models=[*self._find_abstract_models(module), *module_models],
            )
            result[module_import_path] = entity

        for app in loaded_project.apps.get_app_configs():

            def _compat_models_module_is_module_type(
                models_module: Any,
            ) -> types.ModuleType | None:
                # older version of django-stubs typed this wrong
                return models_module  # type: ignore[no-any-return]

            if models_module := _compat_models_module_is_module_type(app.models_module):
                import_path = ImportPath.from_module(models_module)
                if import_path not in result:
                    result[import_path] = self.module_creator(
                        import_path=import_path,
                        module=models_module,
                        models=list(self._find_abstract_models(models_module)),
                    )

        changed: bool = True
        while changed:
            changed = self._find_non_direct_modules(result)

        return result

    def _find_non_direct_modules(
        self, result: dict[protocols.ImportPath, protocols.Module]
    ) -> bool:
        changed: bool = False
        for module in list(result.values()):
            for model in module.defined_models.values():
                for mro_model in model.models_in_mro:
                    if (
                        module_import_path := ImportPath(mro_model.rsplit(".", 1)[0])
                    ) not in result:
                        changed = True
                        module = importlib.import_module(str(module_import_path))
                        result[module_import_path] = self.module_creator(
                            import_path=module_import_path,
                            module=module,
                            models=list(self._find_all_models(module)),
                        )

        return changed

    def _find_all_models(self, module: types.ModuleType) -> Iterator[type[models.Model]]:
        for attr in dir(module):
            val = getattr(module, attr)
            if isinstance(val, type) and issubclass(val.__class__, models.base.ModelBase):
                if val.__module__ != module.__name__:
                    continue

                yield val

    def _find_abstract_models(self, module: types.ModuleType) -> Iterator[type[models.Model]]:
        for model in self._find_all_models(module):
            if hasattr(model, "_meta") and getattr(model._meta, "abstract", False):
                yield model


if TYPE_CHECKING:
    _KMA: protocols.P_InstalledModelsDiscovery = cast(
        DefaultInstalledModulesDiscovery[protocols.P_Project], None
    )



from __future__ import annotations

import contextlib
import pathlib
from collections.abc import Hashable, Iterator, Mapping, Sequence
from typing import TYPE_CHECKING, Any, Literal, NewType, Protocol, TypeVar, Union

from django.apps.registry import Apps
from django.conf import LazySettings
from django.db import models
from typing_extensions import Self

if TYPE_CHECKING:
    from django.contrib.contenttypes.fields import GenericForeignKey
    from django.db.models.fields.related import ForeignObjectRel

T_Project = TypeVar("T_Project", bound="P_Project")
T_VirtualDependency = TypeVar("T_VirtualDependency", bound="P_VirtualDependency")
T_CO_VirtualDependency = TypeVar(
    "T_CO_VirtualDependency", bound="P_VirtualDependency", covariant=True
)
T_COT_VirtualDependency = TypeVar(
    "T_COT_VirtualDependency", bound="P_VirtualDependency", contravariant=True
)

# The difference between Report and ReportUse is that the Report has an api used to build the report
# And ReportUse has no defined api to begin with, and is up to the thing using the report to define
# The api it wants. For this plugin that is defined in extended_mypy_django_plugin._virtual_dependencies
T_Report = TypeVar("T_Report", bound="P_Report")
T_CO_Report = TypeVar("T_CO_Report", bound="P_Report", covariant=True)
T_CO_ReportUse = TypeVar("T_CO_ReportUse", covariant=True)


ImportPath = NewType("ImportPath", str)
FieldsMap = Mapping[str, "Field"]
ModelMap = Mapping[ImportPath, "Model"]
ModelModulesMap = Mapping[ImportPath, "Module"]
ConcreteModelsMap = Mapping[ImportPath, Sequence["Model"]]
SettingsTypesMap = Mapping[str, str]
VirtualDependencyMap = Mapping[ImportPath, T_CO_VirtualDependency]
DjangoField = Union["models.fields.Field[Any, Any]", "ForeignObjectRel", "GenericForeignKey"]


class Hasher(Protocol):
    def __call__(self, *parts: bytes) -> str:
        """
        Given some strings, create a single hash of all that data
        """


class SettingsTypesDiscovery(Protocol[T_Project]):
    """
    Used to discovery the names and types of settings from a loaded project
    """

    def __call__(self, loaded_project: Loaded[T_Project], /) -> SettingsTypesMap: ...


class InstalledModelsDiscovery(Protocol[T_Project]):
    """
    Used to discover installed modules containing Django ORM models in a loaded project
    """

    def __call__(self, loaded_project: Loaded[T_Project], /) -> ModelModulesMap: ...


class ConcreteModelsDiscovery(Protocol[T_Project]):
    """
    Used to determine concrete models
    """

    def __call__(
        self, loaded_project: Loaded[T_Project], all_models: ModelMap, /
    ) -> ConcreteModelsMap: ...


class Discovery(Protocol[T_Project]):
    """
    A container for all the different discovery helpers
    """

    @property
    def discover_settings_types(self) -> SettingsTypesDiscovery[T_Project]:
        """
        Used to discover settings names and their types
        """

    @property
    def discover_installed_models(self) -> InstalledModelsDiscovery[T_Project]:
        """
        Used to discover installed modules containing Django ORM models
        """

    @property
    def discover_concrete_models(self) -> ConcreteModelsDiscovery[T_Project]:
        """
        Used to determine the concrete models for any model
        """


class Project(Protocol):
    """
    Represents a Django project to be analyzed
    """

    @property
    def root_dir(self) -> pathlib.Path:
        """
        Where the django project lives
        """

    @property
    def additional_sys_path(self) -> Sequence[str]:
        """
        Any additional paths that need to be added to sys.path
        """

    @property
    def env_vars(self) -> Mapping[str, str]:
        """
        Any additional environment variables needed to setup Django
        """

    @contextlib.contextmanager
    def setup_sys_path_and_env_vars(self) -> Iterator[None]:
        """
        Do necessary work to setup and cleanup changes to sys.path and os.environ to prepare
        for a django instantiation.
        """

    def load_project(self) -> Loaded[Self]:
        """
        Do necessary work to load Django into memory

        It is expected that an implementation will use self.setup_sys_path_and_env_vars to
        setup and cleanup required changes to sys.path and os.environ
        """


class Loaded(Protocol[T_Project]):
    """
    Represents a Django project that has been setup and loaded into memory
    """

    @property
    def project(self) -> T_Project:
        """
        The project that was loaded
        """

    @property
    def root_dir(self) -> pathlib.Path:
        """
        Where the django project lives
        """

    @property
    def env_vars(self) -> Mapping[str, str]:
        """
        Any additional environment variables that were used during Django setup
        """

    @property
    def settings(self) -> LazySettings:
        """
        The instantiated Django settings object
        """

    @property
    def apps(self) -> Apps:
        """
        The instantiated Django apps registry
        """

    def perform_discovery(self) -> Discovered[T_Project]:
        """
        Perform discovery of important information from the loaded Django project
        """


class Discovered(Protocol[T_Project]):
    @property
    def loaded_project(self) -> Loaded[T_Project]:
        """
        The loaded django project that was analyzed
        """

    @property
    def installed_models_modules(self) -> ModelModulesMap:
        """
        The known modules that contain installed Django Models
        """

    @property
    def all_models(self) -> ModelMap:
        """
        A map of all the models in the project
        """

    @property
    def installed_apps(self) -> list[str]:
        """
        The value of the settings.INSTALLED_APPS setting.
        """

    @property
    def settings_types(self) -> SettingsTypesMap:
        """
        All the django settings and a string representation of their type
        """

    @property
    def concrete_models(self) -> ConcreteModelsMap:
        """
        The map of models to their concrete equivalent
        """


class Module(Protocol, Hashable):
    """
    The models contained within a specific module
    """

    @property
    def import_path(self) -> ImportPath:
        """
        The full import path for this module
        """

    @property
    def defined_models(self) -> ModelMap:
        """
        A map of the installed models defined in this module
        """


class Model(Protocol, Hashable):
    """
    Represents the information contained by a Django model
    """

    @property
    def model_name(self) -> str:
        """
        The name of the class that this model represents
        """

    @property
    def module_import_path(self) -> ImportPath:
        """
        The import path to the module this model lives in
        """

    @property
    def import_path(self) -> ImportPath:
        """
        The full import path to this model
        """

    @property
    def is_abstract(self) -> bool:
        """
        Whether this model is abstract
        """

    @property
    def default_custom_queryset(self) -> ImportPath | None:
        """
        The import path to the default custom queryset for this model if one is defined
        """

    @property
    def all_fields(self) -> FieldsMap:
        """
        The fields associated with this model
        """

    @property
    def models_in_mro(self) -> Sequence[ImportPath]:
        """
        The import paths to the classes in the mro for this model

        As long as they:
        * aren't this model
        * not django.db.models.Model
        * are subclasses of django.db.models.Model
        """


class Field(Protocol):
    """
    Represents a single field on a model
    """

    @property
    def model_import_path(self) -> ImportPath:
        """
        The import path to the model this field is defined on
        """

    @property
    def field_type(self) -> ImportPath:
        """
        The import path to the type of object used to represent this type
        """

    @property
    def related_model(self) -> ImportPath | None:
        """
        The model that is related to this field if the field represents a relationship
        """


class VirtualDependencyNamer(Protocol):
    @property
    def namespace(self) -> ImportPath:
        """
        The namespace for the virtual dependencies
        """

    def __call__(self, module: ImportPath, /) -> ImportPath:
        """
        Return a deterministically determined name representing this module import path
        """


class VirtualDependencyMaker(Protocol[T_Project, T_CO_VirtualDependency]):
    """
    Responsible for generating a virtual dependency
    """

    def __call__(
        self, *, discovered_project: Discovered[T_Project], module: Module
    ) -> T_CO_VirtualDependency: ...


class VirtualDependencyGenerator(Protocol[T_Project, T_CO_VirtualDependency]):
    """
    Object that generates the objects representing the virtual dependencies
    """

    def __call__(
        self, *, discovered_project: Discovered[T_Project]
    ) -> VirtualDependencyMap[T_CO_VirtualDependency]:
        """
        Generate the virtual dependencies for this project
        """


class VirtualDependencyInstaller(Protocol[T_CO_VirtualDependency, T_Report]):
    """
    Used Install the virtual dependencies into their destination.

    Implementations should:

    * Build the reports in the scratch root before installing in the final destination
    * Clear out reports in the final destination that represent modules that don't exist anymore
    """

    def __call__(
        self,
        *,
        scratch_root: pathlib.Path,
        destination: pathlib.Path,
        virtual_namespace: ImportPath,
        report_factory: ReportFactory[T_CO_VirtualDependency, T_Report],
    ) -> CombinedReport[T_Report]: ...


class VirtualDependencySummary(Protocol):
    """
    Represents the different hashes that make up a full hash for a virtual dependency

    This is used to determine if the dependency has changed or not
    """

    @property
    def virtual_namespace(self) -> ImportPath:
        """
        The import path of the virtual dependencies
        """

    @property
    def virtual_import_path(self) -> ImportPath:
        """
        The import path the virtual dependency lives at
        """

    @property
    def module_import_path(self) -> ImportPath:
        """
        The import path to the real module this virtual dependency represents
        """

    @property
    def significant_info(self) -> Sequence[str]:
        """
        Strings representing significant information related to the module

        The idea is that changes in this list should warrant reloading Django in a fresh environment.
        """


class VirtualDependency(Protocol):
    """
    Represents the information held by a virtual dependency for a module
    """

    @property
    def module(self) -> Module:
        """
        The module represented by this virtual dependency
        """

    @property
    def summary(self) -> VirtualDependencySummary:
        """
        The parts that make up a final hash for this virtual dependency
        """

    @property
    def all_related_models(self) -> Sequence[ImportPath]:
        """
        All the models that are related to the module represented by this virtual dependency
        """

    @property
    def concrete_models(self) -> ConcreteModelsMap:
        """
        The models known by this module and their concrete children
        """


class Report(Protocol):
    """
    A report of information that can be used by the mypy plugin to easily get
    information from the virtual dependencies
    """

    def register_module(
        self,
        *,
        module_import_path: ImportPath,
        virtual_import_path: ImportPath,
    ) -> None:
        """
        Register a module to it's virtual path
        """

    def register_model(
        self,
        *,
        model_import_path: ImportPath,
        virtual_import_path: ImportPath,
        concrete_name: str,
        concrete_queryset_name: str,
        concrete_models: Sequence[Model],
    ) -> None:
        """
        Register details about a model
        """


class CombinedReport(Protocol[T_CO_ReportUse]):
    @property
    def version(self) -> str:
        """
        Used to determine a string that represents the state of the report
        such that the string changing indicates the state of the project changed
        """

    @property
    def report(self) -> T_CO_ReportUse:
        """
        The final combined report
        """

    def ensure_virtual_dependency(self, *, module_import_path: str) -> None:
        """
        Ensure this module has a virtual dependency if it should have one

        This is important because if a module contains models but isn't installed then when it is
        added to settings.INSTALLED_APPS without any changes, then we can't retrospectively give it
        a virtual dependency in get_additional_deps
        """


class ReportMaker(Protocol[T_CO_Report]):
    """
    Used to construct a report
    """

    def __call__(self) -> T_CO_Report: ...


class RenderedVirtualDependency(Protocol[T_CO_Report]):
    @property
    def content(self) -> str:
        """
        The string representing the content of the virtual dependency
        """

    @property
    def summary_hash(self) -> str | None:
        """
        A hash representing the state of the dependency

        This should be None if the module is not installed
        """

    @property
    def report(self) -> T_CO_Report:
        """
        The report representing the information in the content
        """

    @property
    def virtual_import_path(self) -> ImportPath:
        """
        The import path to this virtual dependency
        """


class VirtualDependencyScribe(Protocol[T_COT_VirtualDependency, T_CO_Report]):
    """
    Used to render the on disk representation of a virtual dependency along with the information
    contained in that dependency
    """

    def __call__(
        self,
        *,
        virtual_dependency: T_COT_VirtualDependency,
        all_virtual_dependencies: VirtualDependencyMap[T_COT_VirtualDependency],
    ) -> RenderedVirtualDependency[T_CO_Report]: ...


class MakeEmptyVirtualDepContent(Protocol):
    """
    Used to make the content that would go into a virtual dependency for an uninstalled module
    """

    def __call__(self, *, module_import_path: ImportPath) -> str: ...


class EmptyVirtualDepWriter(Protocol):
    """
    Used to generate the on disk representation of a virtual dependency for a module that has
    been determined as an uninstalled module with django models in it.

    It should return the import path for the created virtual import, or None if there was an
    existing virtual dependency or none was made
    """

    def __call__(self, *, module_import_path: ImportPath) -> ImportPath | None: ...


class ReportInstaller(Protocol):
    """
    Used to write reports to the file system
    """

    def write_report(
        self,
        *,
        scratch_root: pathlib.Path,
        virtual_import_path: ImportPath,
        content: str,
        summary_hash: str | Literal[False] | None,
    ) -> bool:
        """
        Write a single report to the scratch path

        If summary_hash is False, then only write if the virtual dep doesn't already exist

        Return whether a report was written
        """

    def install_reports(
        self,
        *,
        scratch_root: pathlib.Path,
        destination: pathlib.Path,
        virtual_namespace: ImportPath,
    ) -> None:
        """
        Copy reports from scratch_root into the destination when the reports on the destination
        are different to the reports in the scratch path.

        Also, delete redundant reports from destination
        """


class ReportCombiner(Protocol[T_CO_Report]):
    """
    Used to combine many reports into one
    """

    @property
    def reports(self) -> Sequence[T_CO_Report]:
        """
        The reports to combine
        """

    def combine(
        self, *, version: str, write_empty_virtual_dep: EmptyVirtualDepWriter
    ) -> CombinedReport[T_CO_Report]:
        """
        Return a single report that represents all the provided reports as one
        """


class ReportCombinerMaker(Protocol[T_Report]):
    """
    Used to create a report combiner
    """

    def __call__(self, *, reports: Sequence[T_Report]) -> ReportCombiner[T_Report]: ...


class ReportFactory(Protocol[T_COT_VirtualDependency, T_Report]):
    """
    Holds everything necessary for creating reports
    """

    @property
    def report_installer(self) -> ReportInstaller: ...

    @property
    def report_maker(self) -> ReportMaker[T_Report]: ...

    @property
    def report_combiner_maker(self) -> ReportCombinerMaker[T_Report]: ...

    @property
    def make_empty_virtual_dependency_content(self) -> MakeEmptyVirtualDepContent: ...

    def deploy_scribes(
        self, virtual_dependencies: VirtualDependencyMap[T_COT_VirtualDependency]
    ) -> Iterator[RenderedVirtualDependency[T_Report]]: ...

    def determine_version(
        self,
        *,
        destination: pathlib.Path,
        virtual_namespace: ImportPath,
        project_version: str,
        written_dependencies: Sequence[RenderedVirtualDependency[T_Report]],
    ) -> str:
        """
        Determine a version for the project, such that it only changes if the mypy daemon needs to be restarted
        """


class VirtualDependencyHandler(Protocol[T_CO_ReportUse]):
    """
    This is the interface required by the mypy plugin to create virtual dependencies and get a report
    of the information held by those virtual dependencies
    """

    def __call__(
        self,
        *,
        project_root: pathlib.Path,
        django_settings_module: str,
        virtual_deps_destination: pathlib.Path,
    ) -> CombinedReport[T_CO_ReportUse]: ...


if TYPE_CHECKING:
    P_Model = Model
    P_Field = Field
    P_Module = Module
    P_Hasher = Hasher
    P_Project = Project
    P_Loaded = Loaded[P_Project]
    P_Discovery = Discovery[P_Project]
    P_Discovered = Discovered[P_Project]
    P_SettingsTypesDiscovery = SettingsTypesDiscovery[P_Project]
    P_ConcreteModelsDiscovery = ConcreteModelsDiscovery[P_Project]
    P_InstalledModelsDiscovery = InstalledModelsDiscovery[P_Project]

    P_Report = Report
    P_VirtualDependency = VirtualDependency

    P_ReportMaker = ReportMaker[P_Report]
    P_ReportFactory = ReportFactory[P_VirtualDependency, P_Report]
    P_CombinedReport = CombinedReport[P_Report]
    P_ReportInstaller = ReportInstaller
    P_ReportCombiner = ReportCombiner[P_Report]
    P_ReportCombinerMaker = ReportCombinerMaker[P_Report]

    P_EmptyVirtualDepWriter = EmptyVirtualDepWriter
    P_MakeEmptyVirtualDepContent = MakeEmptyVirtualDepContent

    P_VirtualDependencyMaker = VirtualDependencyMaker[P_Project, P_VirtualDependency]
    P_VirtualDependencyNamer = VirtualDependencyNamer
    P_VirtualDependencyGenerator = VirtualDependencyGenerator[P_Project, P_VirtualDependency]
    P_VirtualDependencySummary = VirtualDependencySummary
    P_VirtualDependencyInstaller = VirtualDependencyInstaller[P_VirtualDependency, P_Report]
    P_VirtualDependencyScribe = VirtualDependencyScribe[P_VirtualDependency, P_Report]
    P_RenderedVirtualDependency = RenderedVirtualDependency[P_Report]
    P_VirtualDependencyHandler = VirtualDependencyHandler[P_Report]



from mypy.plugin import AnalyzeTypeContext
from mypy.types import Type as MypyType
from mypy.types import TypeQuery, TypeVarType, get_proper_type

from . import protocols


class HasTypeVars(TypeQuery[bool]):
    """
    Find where we have a concrete annotation
    """

    def __init__(self) -> None:
        super().__init__(any)

    def visit_type_var(self, t: TypeVarType) -> bool:
        return True


class Analyzer:
    def __init__(self, make_resolver: protocols.ResolverMaker) -> None:
        self.make_resolver = make_resolver

    def analyze_type(
        self, ctx: AnalyzeTypeContext, annotation: protocols.KnownAnnotations
    ) -> MypyType:
        """
        We resolve annotations at this point. Unless the type being analyzed involves type vars.

        Resolving type vars requires we wait until we are analyzing method/function calls. Between now
        and then we replace the type with an unbound type that wraps a resolved instance because when we
        can resolve the type vars we can't resolve what the type var actually is!
        """
        args = ctx.type.args
        if len(args) != 1:
            ctx.api.fail("Concrete annotations must contain exactly one argument", ctx.context)
            return ctx.type

        model_type = get_proper_type(ctx.api.analyze_type(args[0]))

        resolver = self.make_resolver(ctx=ctx)

        if model_type.accept(HasTypeVars()):
            ctx.api.fail(
                "Using a concrete annotation on a TypeVar is not currently supported", ctx.context
            )
            return ctx.type

        resolved = resolver.resolve(annotation, model_type)
        if resolved is None:
            return ctx.type
        else:
            return resolved



import dataclasses
import functools
import pathlib
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Literal, cast

import pytest

from extended_mypy_django_plugin.django_analysis import (
    ImportPath,
    Module,
    Project,
    adler32_hash,
    protocols,
    virtual_dependencies,
)


class TestVirtualDependencyGenerator:
    def test_it_can_generate_virtual_dependencies(
        self, discovered_django_example: protocols.Discovered[Project]
    ) -> None:
        class CustomVirtualDependency(virtual_dependencies.VirtualDependency[Project]):
            @classmethod
            def find_significant_info_from_module(
                cls,
                *,
                discovered_project: protocols.Discovered[Project],
                module: protocols.Module,
                concrete_models: protocols.ConcreteModelsMap,
            ) -> Iterator[str]:
                yield f"__significant__{module.import_path}__"

        virtual_dependency_namer = virtual_dependencies.VirtualDependencyNamer(
            namespace=ImportPath("__virtual__"), hasher=adler32_hash
        )

        virtual_dependency_maker = functools.partial(
            CustomVirtualDependency.create, virtual_dependency_namer=virtual_dependency_namer
        )

        generated = virtual_dependencies.VirtualDependencyGenerator(
            virtual_dependency_maker=virtual_dependency_maker
        )(discovered_project=discovered_django_example)

        def IsModule(import_path: str) -> protocols.Module:
            return discovered_django_example.installed_models_modules[ImportPath(import_path)]

        def IsModel(import_path: str) -> protocols.Model:
            return discovered_django_example.all_models[ImportPath(import_path)]

        assert generated == {
            ImportPath("django.contrib.admin.models"): CustomVirtualDependency(
                module=IsModule("django.contrib.admin.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_2456226428"),
                    module_import_path=ImportPath("django.contrib.admin.models"),
                    significant_info=["__significant__django.contrib.admin.models__"],
                ),
                all_related_models=[
                    ImportPath("django.contrib.admin.models.LogEntry"),
                    ImportPath("django.contrib.auth.models.User"),
                    ImportPath("django.contrib.contenttypes.models.ContentType"),
                ],
                concrete_models={
                    ImportPath("django.contrib.admin.models.LogEntry"): [
                        IsModel("django.contrib.admin.models.LogEntry")
                    ]
                },
            ),
            ImportPath("django.contrib.auth.base_user"): CustomVirtualDependency(
                module=IsModule("django.contrib.auth.base_user"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_2833058650"),
                    module_import_path=ImportPath("django.contrib.auth.base_user"),
                    significant_info=["__significant__django.contrib.auth.base_user__"],
                ),
                all_related_models=[
                    ImportPath("django.contrib.auth.base_user.AbstractBaseUser"),
                ],
                concrete_models={
                    ImportPath("django.contrib.auth.base_user.AbstractBaseUser"): [
                        IsModel("django.contrib.auth.models.User")
                    ],
                },
            ),
            ImportPath("django.contrib.auth.models"): CustomVirtualDependency(
                module=IsModule("django.contrib.auth.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_2289830437"),
                    module_import_path=ImportPath("django.contrib.auth.models"),
                    significant_info=["__significant__django.contrib.auth.models__"],
                ),
                all_related_models=[
                    ImportPath("django.contrib.admin.models.LogEntry"),
                    ImportPath("django.contrib.auth.models.AbstractUser"),
                    ImportPath("django.contrib.auth.models.Group"),
                    ImportPath("django.contrib.auth.models.Permission"),
                    ImportPath("django.contrib.auth.models.PermissionsMixin"),
                    ImportPath("django.contrib.auth.models.User"),
                    ImportPath("django.contrib.contenttypes.models.ContentType"),
                ],
                concrete_models={
                    ImportPath("django.contrib.auth.models.AbstractUser"): [
                        IsModel("django.contrib.auth.models.User")
                    ],
                    ImportPath("django.contrib.auth.models.Group"): [
                        IsModel("django.contrib.auth.models.Group")
                    ],
                    ImportPath("django.contrib.auth.models.Permission"): [
                        IsModel("django.contrib.auth.models.Permission")
                    ],
                    ImportPath("django.contrib.auth.models.PermissionsMixin"): [
                        IsModel("django.contrib.auth.models.User")
                    ],
                    ImportPath("django.contrib.auth.models.User"): [
                        IsModel("django.contrib.auth.models.User")
                    ],
                },
            ),
            ImportPath("django.contrib.contenttypes.models"): CustomVirtualDependency(
                module=IsModule("django.contrib.contenttypes.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_3961720227"),
                    module_import_path=ImportPath("django.contrib.contenttypes.models"),
                    significant_info=["__significant__django.contrib.contenttypes.models__"],
                ),
                all_related_models=[
                    ImportPath("django.contrib.admin.models.LogEntry"),
                    ImportPath("django.contrib.auth.models.Permission"),
                    ImportPath("django.contrib.contenttypes.models.ContentType"),
                ],
                concrete_models={
                    ImportPath("django.contrib.contenttypes.models.ContentType"): [
                        IsModel("django.contrib.contenttypes.models.ContentType")
                    ]
                },
            ),
            ImportPath("django.contrib.sessions.base_session"): CustomVirtualDependency(
                module=IsModule("django.contrib.sessions.base_session"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_113708644"),
                    module_import_path=ImportPath("django.contrib.sessions.base_session"),
                    significant_info=["__significant__django.contrib.sessions.base_session__"],
                ),
                all_related_models=[
                    ImportPath("django.contrib.sessions.base_session.AbstractBaseSession"),
                ],
                concrete_models={
                    ImportPath("django.contrib.sessions.base_session.AbstractBaseSession"): [
                        IsModel("django.contrib.sessions.models.Session")
                    ]
                },
            ),
            ImportPath("django.contrib.sessions.models"): CustomVirtualDependency(
                module=IsModule("django.contrib.sessions.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_3074165738"),
                    module_import_path=ImportPath("django.contrib.sessions.models"),
                    significant_info=["__significant__django.contrib.sessions.models__"],
                ),
                all_related_models=[
                    ImportPath("django.contrib.sessions.models.Session"),
                ],
                concrete_models={
                    ImportPath("django.contrib.sessions.models.Session"): [
                        IsModel("django.contrib.sessions.models.Session")
                    ],
                },
            ),
            ImportPath("djangoexample.exampleapp.models"): CustomVirtualDependency(
                module=IsModule("djangoexample.exampleapp.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_3347844205"),
                    module_import_path=ImportPath("djangoexample.exampleapp.models"),
                    significant_info=["__significant__djangoexample.exampleapp.models__"],
                ),
                all_related_models=[
                    ImportPath("djangoexample.exampleapp.models.Child1"),
                    ImportPath("djangoexample.exampleapp.models.Child2"),
                    ImportPath("djangoexample.exampleapp.models.Child3"),
                    ImportPath("djangoexample.exampleapp.models.Child4"),
                    ImportPath("djangoexample.exampleapp.models.Parent"),
                    ImportPath("djangoexample.exampleapp.models.Parent2"),
                ],
                concrete_models={
                    ImportPath("djangoexample.exampleapp.models.Child1"): [
                        IsModel("djangoexample.exampleapp.models.Child1"),
                    ],
                    ImportPath("djangoexample.exampleapp.models.Child2"): [
                        IsModel("djangoexample.exampleapp.models.Child2")
                    ],
                    ImportPath("djangoexample.exampleapp.models.Child3"): [
                        IsModel("djangoexample.exampleapp.models.Child3")
                    ],
                    ImportPath("djangoexample.exampleapp.models.Child4"): [
                        IsModel("djangoexample.exampleapp.models.Child4")
                    ],
                    ImportPath("djangoexample.exampleapp.models.Parent"): [
                        IsModel("djangoexample.exampleapp.models.Child1"),
                        IsModel("djangoexample.exampleapp.models.Child2"),
                        IsModel("djangoexample.exampleapp.models.Child3"),
                        IsModel("djangoexample.exampleapp.models.Child4"),
                        IsModel("djangoexample.exampleapp2.models.ChildOther"),
                        IsModel("djangoexample.exampleapp2.models.ChildOther2"),
                    ],
                    ImportPath("djangoexample.exampleapp.models.Parent2"): [
                        IsModel("djangoexample.exampleapp.models.Child3"),
                        IsModel("djangoexample.exampleapp.models.Child4"),
                    ],
                },
            ),
            ImportPath("djangoexample.exampleapp2.models"): CustomVirtualDependency(
                module=IsModule("djangoexample.exampleapp2.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_3537308831"),
                    module_import_path=ImportPath("djangoexample.exampleapp2.models"),
                    significant_info=["__significant__djangoexample.exampleapp2.models__"],
                ),
                all_related_models=[
                    ImportPath("djangoexample.exampleapp2.models.ChildOther"),
                    ImportPath("djangoexample.exampleapp2.models.ChildOther2"),
                ],
                concrete_models={
                    ImportPath("djangoexample.exampleapp2.models.ChildOther"): [
                        IsModel("djangoexample.exampleapp2.models.ChildOther")
                    ],
                    ImportPath("djangoexample.exampleapp2.models.ChildOther2"): [
                        IsModel("djangoexample.exampleapp2.models.ChildOther2"),
                    ],
                },
            ),
            ImportPath("djangoexample.only_abstract.models"): CustomVirtualDependency(
                module=IsModule("djangoexample.only_abstract.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_4035906997"),
                    module_import_path=ImportPath("djangoexample.only_abstract.models"),
                    significant_info=["__significant__djangoexample.only_abstract.models__"],
                ),
                all_related_models=[
                    ImportPath("djangoexample.only_abstract.models.AnAbstract"),
                ],
                concrete_models={
                    ImportPath("djangoexample.only_abstract.models.AnAbstract"): [],
                },
            ),
            ImportPath("djangoexample.relations1.models"): CustomVirtualDependency(
                module=IsModule("djangoexample.relations1.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_3327724610"),
                    module_import_path=ImportPath("djangoexample.relations1.models"),
                    significant_info=["__significant__djangoexample.relations1.models__"],
                ),
                all_related_models=[
                    ImportPath("djangoexample.relations1.models.Abstract"),
                    ImportPath("djangoexample.relations1.models.Child1"),
                    ImportPath("djangoexample.relations1.models.Child2"),
                    ImportPath("djangoexample.relations1.models.Concrete1"),
                    ImportPath("djangoexample.relations1.models.Concrete2"),
                    ImportPath("djangoexample.relations2.models.Thing"),
                ],
                concrete_models={
                    ImportPath("djangoexample.relations1.models.Abstract"): [
                        IsModel("djangoexample.relations1.models.Child1"),
                        IsModel("djangoexample.relations1.models.Child2"),
                    ],
                    ImportPath("djangoexample.relations1.models.Child1"): [
                        IsModel("djangoexample.relations1.models.Child1"),
                    ],
                    ImportPath("djangoexample.relations1.models.Child2"): [
                        IsModel("djangoexample.relations1.models.Child2"),
                    ],
                    ImportPath("djangoexample.relations1.models.Concrete1"): [
                        IsModel("djangoexample.relations1.models.Concrete1"),
                    ],
                    ImportPath("djangoexample.relations1.models.Concrete2"): [
                        IsModel("djangoexample.relations1.models.Concrete2"),
                    ],
                },
            ),
            ImportPath("djangoexample.relations2.models"): CustomVirtualDependency(
                module=IsModule("djangoexample.relations2.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_3328248899"),
                    module_import_path=ImportPath("djangoexample.relations2.models"),
                    significant_info=["__significant__djangoexample.relations2.models__"],
                ),
                all_related_models=[
                    ImportPath("djangoexample.relations1.models.Concrete1"),
                    ImportPath("djangoexample.relations2.models.Thing"),
                ],
                concrete_models={
                    ImportPath("djangoexample.relations2.models.Thing"): [
                        IsModel("djangoexample.relations2.models.Thing")
                    ],
                },
            ),
            ImportPath("djangoexample.empty_models.models"): CustomVirtualDependency(
                module=IsModule("djangoexample.empty_models.models"),
                summary=virtual_dependencies.VirtualDependencySummary(
                    virtual_namespace=ImportPath("__virtual__"),
                    virtual_import_path=ImportPath("__virtual__.mod_3808300370"),
                    module_import_path=ImportPath("djangoexample.empty_models.models"),
                    significant_info=["__significant__djangoexample.empty_models.models__"],
                ),
                all_related_models=[],
                concrete_models={},
            ),
        }


class TestVirtualDependencyInstaller:
    def test_it_uses_the_report_factory(self, tmp_path_factory: pytest.TempPathFactory) -> None:
        scratch_root = tmp_path_factory.mktemp("scratch_root")
        destination = tmp_path_factory.mktemp("destination")

        installed: list[tuple[pathlib.Path, pathlib.Path, protocols.ImportPath]] = []
        written: dict[
            tuple[pathlib.Path, protocols.ImportPath], tuple[str, str | Literal[False] | None]
        ] = {}

        virtual_dependency_namer = virtual_dependencies.VirtualDependencyNamer(
            namespace=ImportPath("__virtual__"), hasher=adler32_hash
        )

        class Dep(virtual_dependencies.VirtualDependency[Project]):
            pass

        @dataclasses.dataclass
        class Report:
            modules: set[tuple[protocols.ImportPath, protocols.ImportPath]] = dataclasses.field(
                default_factory=set
            )
            combined: bool = False

            def register_module(
                self,
                *,
                module_import_path: protocols.ImportPath,
                virtual_import_path: protocols.ImportPath,
            ) -> None:
                self.modules.add((module_import_path, virtual_import_path))

            def register_model(
                self,
                *,
                model_import_path: protocols.ImportPath,
                virtual_import_path: protocols.ImportPath,
                concrete_name: str,
                concrete_queryset_name: str,
                concrete_models: Sequence[protocols.Model],
            ) -> None:
                raise ValueError("not called")

        @dataclasses.dataclass
        class ReportCombiner:
            reports: Sequence[Report]

            def combine(
                self, *, version: str, write_empty_virtual_dep: protocols.EmptyVirtualDepWriter
            ) -> protocols.CombinedReport[Report]:
                final = Report(combined=True)
                for report in self.reports:
                    final.modules |= report.modules

                return virtual_dependencies.CombinedReport(
                    version=version, report=final, write_empty_virtual_dep=write_empty_virtual_dep
                )

        @dataclasses.dataclass
        class ReportInstaller:
            def write_report(
                self,
                *,
                scratch_root: pathlib.Path,
                summary_hash: str | Literal[False] | None,
                virtual_import_path: protocols.ImportPath,
                content: str,
            ) -> bool:
                key = (scratch_root, virtual_import_path)
                exists = any(v == virtual_import_path for _, v in written)
                if exists and summary_hash is False:
                    return False
                assert not exists
                written[key] = (content, summary_hash)
                return True

            def install_reports(
                self,
                *,
                scratch_root: pathlib.Path,
                destination: pathlib.Path,
                virtual_namespace: protocols.ImportPath,
            ) -> None:
                installed.append((scratch_root, destination, virtual_namespace))

        class ReportFactory:
            def __init__(self) -> None:
                self.report_installer = ReportInstaller()
                self.report_maker = Report
                self.report_combiner_maker = ReportCombiner

            def deploy_scribes(
                self, all_virtual_dependencies: protocols.VirtualDependencyMap[Dep]
            ) -> Iterator[protocols.RenderedVirtualDependency[Report]]:
                for virtual_dependency in all_virtual_dependencies.values():
                    report = self.report_maker()
                    report.register_module(
                        module_import_path=virtual_dependency.summary.module_import_path,
                        virtual_import_path=virtual_dependency.summary.virtual_import_path,
                    )
                    yield virtual_dependencies.RenderedVirtualDependency(
                        content=f"CONTENT__{virtual_dependency.summary.module_import_path}",
                        summary_hash=f"SUMMARY__{virtual_dependency.summary.module_import_path}",
                        report=report,
                        virtual_import_path=virtual_dependency.summary.virtual_import_path,
                    )

            def make_empty_virtual_dependency_content(
                self, *, module_import_path: protocols.ImportPath
            ) -> str:
                return "empty"

            def determine_version(
                self,
                *,
                destination: pathlib.Path,
                virtual_namespace: protocols.ImportPath,
                project_version: str,
                written_dependencies: Sequence[
                    protocols.RenderedVirtualDependency[protocols.T_Report]
                ],
            ) -> str:
                assert project_version == "__project_version__"
                assert virtual_namespace == ImportPath("__virtual__")
                return "__version__"

        if TYPE_CHECKING:
            _D: protocols.VirtualDependency = cast(Dep, None)
            _R: protocols.Report = cast(Report, None)
            _RC: protocols.ReportCombiner[Report] = cast(ReportCombiner, None)
            _RCM: protocols.ReportCombinerMaker[Report] = ReportCombiner
            _RI: protocols.ReportInstaller = cast(ReportInstaller, None)
            _RF: protocols.ReportFactory[Dep, Report] = cast(ReportFactory, None)

        installer = virtual_dependencies.VirtualDependencyInstaller[Dep, Report](
            project_version="__project_version__",
            virtual_dependency_namer=virtual_dependency_namer,
            virtual_dependencies={
                ImportPath("M1.models"): Dep(
                    module=Module(import_path=ImportPath("M1.models"), defined_models={}),
                    summary=virtual_dependencies.VirtualDependencySummary(
                        virtual_namespace=ImportPath("__virtual__"),
                        virtual_import_path=ImportPath("__virtual__.mod_239797041"),
                        module_import_path=ImportPath("M1"),
                        significant_info=["__significant__django.contrib.admin.models__"],
                    ),
                    all_related_models=[],
                    concrete_models={},
                ),
                ImportPath("M2"): Dep(
                    module=Module(import_path=ImportPath("M2"), defined_models={}),
                    summary=virtual_dependencies.VirtualDependencySummary(
                        virtual_namespace=ImportPath("__virtual__"),
                        virtual_import_path=ImportPath("__virtual__.M2"),
                        module_import_path=ImportPath("M2"),
                        significant_info=["__significant__django.contrib.admin.models__"],
                    ),
                    all_related_models=[],
                    concrete_models={},
                ),
            },
        )

        assert written == {}
        assert installed == []
        report = installer(
            scratch_root=scratch_root,
            destination=destination,
            virtual_namespace=ImportPath("__virtual__"),
            report_factory=ReportFactory(),
        )

        assert report.version == "__version__"
        assert report.report == Report(
            combined=True,
            modules={
                (ImportPath("M1"), ImportPath("__virtual__.mod_239797041")),
                (ImportPath("M2"), ImportPath("__virtual__.M2")),
            },
        )

        assert written == {
            (scratch_root, ImportPath("__virtual__.mod_239797041")): (
                "CONTENT__M1",
                "SUMMARY__M1",
            ),
            (scratch_root, ImportPath("__virtual__.M2")): ("CONTENT__M2", "SUMMARY__M2"),
        }
        assert installed == [(scratch_root, destination, ImportPath("__virtual__"))]

        report.ensure_virtual_dependency(module_import_path="E1.models")
        assert (
            ImportPath("E1.models"),
            ImportPath("__virtual__.mod_235078441"),
        ) in report.report.modules
        assert written == {
            (scratch_root, ImportPath("__virtual__.mod_239797041")): (
                "CONTENT__M1",
                "SUMMARY__M1",
            ),
            (scratch_root, ImportPath("__virtual__.M2")): ("CONTENT__M2", "SUMMARY__M2"),
            # The empty dep is written directly to the final destination
            # cause it happens at mypy time
            (destination, ImportPath("__virtual__.mod_235078441")): ("empty", False),
        }

        # Existing modules are not overridden
        report.ensure_virtual_dependency(module_import_path="M1.models")
        assert written == {
            # We want the content for M1 to have not been changed in destination
            (scratch_root, ImportPath("__virtual__.mod_239797041")): (
                "CONTENT__M1",
                "SUMMARY__M1",
            ),
            (scratch_root, ImportPath("__virtual__.M2")): ("CONTENT__M2", "SUMMARY__M2"),
            (destination, ImportPath("__virtual__.mod_235078441")): ("empty", False),
        }
        assert report.report.modules == {
            (ImportPath("M1"), ImportPath("__virtual__.mod_239797041")),
            (ImportPath("M2"), ImportPath("__virtual__.M2")),
            (ImportPath("E1.models"), ImportPath("__virtual__.mod_235078441")),
        }



from extended_mypy_django_plugin.django_analysis import (
    ImportPath,
    adler32_hash,
    virtual_dependencies,
)


class TestVirtualDependencyNamer:
    def test_it_can_name_virtual_dependencies(self) -> None:
        namer = virtual_dependencies.VirtualDependencyNamer(
            namespace=ImportPath("some.where.nice"), hasher=adler32_hash
        )

        assert namer(ImportPath("my.nice.model")) == "some.where.nice.mod_577176819"
        assert namer(ImportPath("my.other.model")) == "some.where.nice.mod_685704566"

        def bad_hasher(*vals: bytes) -> str:
            return f"__hashed__{b'_'.join(vals).replace(b'.', b'DD').decode()}"

        namer = virtual_dependencies.VirtualDependencyNamer(
            namespace=ImportPath("a.bad.place"), hasher=bad_hasher
        )

        assert namer(ImportPath("my.nice.model")) == "a.bad.place.mod___hashed__myDDniceDDmodel"
        assert namer(ImportPath("my.other.model")) == "a.bad.place.mod___hashed__myDDotherDDmodel"



import pathlib

import pytest

from extended_mypy_django_plugin.django_analysis import Project, discovery, protocols

root_dir = pathlib.Path(__file__).parent.parent.parent


@pytest.fixture(scope="session")
def loaded_django_example() -> protocols.Loaded[Project]:
    return Project(
        root_dir=root_dir / "example",
        additional_sys_path=[str(root_dir)],
        discovery=discovery.Discovery(),
        env_vars={"DJANGO_SETTINGS_MODULE": "djangoexample.settings"},
    ).load_project()


@pytest.fixture(scope="session")
def discovered_django_example(
    loaded_django_example: protocols.Loaded[Project],
) -> protocols.Discovered[Project]:
    return loaded_django_example.perform_discovery()



import functools
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, cast

from mypy.nodes import Context, PlaceholderNode, TypeAlias, TypeInfo
from mypy.plugin import (
    AnalyzeTypeContext,
    AttributeContext,
    DynamicClassDefContext,
    FunctionContext,
    MethodContext,
)
from mypy.semanal import SemanticAnalyzer
from mypy.typeanal import TypeAnalyser
from mypy.types import (
    AnyType,
    Instance,
    PlaceholderType,
    ProperType,
    TypeOfAny,
    TypeType,
    UnionType,
    get_proper_type,
)
from mypy.types import Type as MypyType
from typing_extensions import Self, assert_never

from . import protocols


class ShouldDefer(Exception):
    pass


class FailedLookup(Exception):
    pass


class AnnotationResolver:
    @classmethod
    def create(
        cls,
        *,
        get_concrete_aliases: protocols.AliasGetter,
        get_queryset_aliases: protocols.AliasGetter,
        plugin_lookup_fully_qualified: protocols.LookupFullyQualified,
        ctx: protocols.ValidContextForAnnotationResolver,
    ) -> Self:
        """
        This classmethod constructor lets us normalise the ctx to satisfy the interface the
        AnnotationResolver expects.

        Because each ctx type has a different api on it that provides a different set of
        abilities.
        """

        def sem_defer(sem_api: SemanticAnalyzer) -> bool:
            """
            The semantic analyzer is the only api that can actually defer

            Return True if was able to defer
            """
            if sem_api.final_iteration:
                return False
            else:
                sem_api.defer()
                return True

        def _lookup_info(sem_api: SemanticAnalyzer | None, fullname: str) -> TypeInfo | None:
            """
            If we have the semantic api, there's more we can do when trying to lookup
            some name
            """
            if sem_api is not None:
                instance = sem_api.named_type_or_none(fullname)
                if instance:
                    return instance.type

            sym = plugin_lookup_fully_qualified(fullname)
            if not sym or not isinstance(node := sym.node, TypeInfo):
                return None
            else:
                return node

        def checker_named_type_or_none(
            fullname: str, args: list[MypyType] | None = None
        ) -> Instance | None:
            """
            When we have a TypeChecker we need to replicate close to what the semantic api
            does for named_type_or_none
            """
            sym = plugin_lookup_fully_qualified(fullname)
            if not sym or not isinstance(node := sym.node, TypeInfo):
                return None
            if args:
                return Instance(node, args)
            return Instance(node, [AnyType(TypeOfAny.special_form)] * len(node.defn.type_vars))

        def _lookup_alias(line: int, alias: str) -> Iterator[Instance | PlaceholderType]:
            """
            This is the same regardless of which ctx we have
            """
            try:
                sym = plugin_lookup_fully_qualified(alias)
            except AssertionError:
                raise FailedLookup(f"Failed to lookup {alias}")

            if not sym or isinstance(sym.node, PlaceholderNode):
                yield PlaceholderType(alias, [], line)
                return

            assert sym and isinstance(sym.node, TypeAlias)
            target = get_proper_type(sym.node.target)

            if isinstance(target, Instance):
                yield target
            elif isinstance(target, UnionType):
                for item in target.items:
                    found = get_proper_type(item)
                    assert isinstance(found, Instance | PlaceholderType)
                    yield found
            else:
                raise FailedLookup(f"Expected only an instance or union for {alias}: got {target}")

        fail: protocols.FailFunc
        defer: protocols.DeferFunc
        context: Context
        lookup_info: protocols.LookupInfo
        named_type_or_none: protocols.NamedTypeOrNone

        match ctx:
            case DynamicClassDefContext(api=api):
                assert isinstance(api, SemanticAnalyzer)
                context = ctx.call
                sem_api = api
                defer = functools.partial(sem_defer, sem_api)
                fail = functools.partial(sem_api.fail, ctx=context)
                lookup_info = functools.partial(_lookup_info, sem_api)
                lookup_alias = functools.partial(_lookup_alias, context.line)
                named_type_or_none = sem_api.named_type_or_none
            case AnalyzeTypeContext(api=api):
                assert isinstance(api, TypeAnalyser)
                assert isinstance(api.api, SemanticAnalyzer)
                context = ctx.context
                sem_api = api.api
                defer = functools.partial(sem_defer, sem_api)
                fail = functools.partial(sem_api.fail, ctx=context)
                lookup_info = functools.partial(_lookup_info, sem_api)
                lookup_alias = functools.partial(_lookup_alias, context.line)
                named_type_or_none = sem_api.named_type_or_none
            case AttributeContext(api=api) | MethodContext(api=api) | FunctionContext(api=api):
                context = ctx.context
                defer = lambda: False
                # The interface for the type checker says fail takes ctx
                # But the implementation of TypeChecker has it as context
                fail = functools.partial(api.fail, context=context)
                lookup_info = functools.partial(_lookup_info, None)
                lookup_alias = functools.partial(_lookup_alias, context.line)
                named_type_or_none = checker_named_type_or_none
            case _:
                assert_never(ctx)

        return cls(
            context=context,
            get_concrete_aliases=get_concrete_aliases,
            get_queryset_aliases=get_queryset_aliases,
            defer=defer,
            fail=fail,
            lookup_info=lookup_info,
            lookup_alias=lookup_alias,
            named_type_or_none=named_type_or_none,
        )

    def __init__(
        self,
        *,
        context: Context,
        get_concrete_aliases: protocols.AliasGetter,
        get_queryset_aliases: protocols.AliasGetter,
        fail: protocols.FailFunc,
        defer: protocols.DeferFunc,
        lookup_alias: protocols.LookupAlias,
        lookup_info: protocols.LookupInfo,
        named_type_or_none: protocols.NamedTypeOrNone,
    ) -> None:
        self._defer = defer
        self._named_type_or_none = named_type_or_none
        self.fail = fail
        self.context = context
        self.lookup_info = lookup_info
        self.lookup_alias = lookup_alias
        self.get_concrete_aliases = get_concrete_aliases
        self.get_queryset_aliases = get_queryset_aliases

    def _flatten_union(self, typ: ProperType) -> Iterator[ProperType]:
        """
        Recursively flatten a union
        """
        if isinstance(typ, UnionType):
            for item in typ.items:
                yield from self._flatten_union(get_proper_type(item))
        else:
            yield typ

    def _concrete_for(
        self, model_type: ProperType, get_aliases: protocols.AliasGetter
    ) -> Instance | TypeType | UnionType | PlaceholderType | None:
        """
        Given some type that represents a model, and an alias getter, determine an
        Instance if we can from the model_type and use it with the alias getter.
        """
        is_type: bool = False

        found: ProperType = model_type
        if isinstance(model_type, TypeType):
            is_type = True
            found = model_type.item

        if isinstance(found, AnyType):
            did_defer = self._defer()
            if not did_defer:
                self.fail("Tried to use concrete annotations on a typing.Any")
            return None

        if not isinstance(found, Instance | UnionType):
            return None

        all_types = list(self._flatten_union(found))
        are_all_instances: bool = True
        names: list[str] = []
        concrete: list[Instance | PlaceholderType] = []

        for item in all_types:
            if not isinstance(item, Instance):
                self.fail(
                    f"Expected to operate on specific classes, got a {item.__class__.__name__}: {item}"
                )
                are_all_instances = False
                continue

            name = item.type.fullname
            names.append(name)
            concrete.extend(self._instances_from_aliases(get_aliases, name))

        if not are_all_instances:
            return None

        if not concrete:
            # We found instances, but couldn't get aliases
            # Either defer and we'll try again later or fail if we can't defer
            did_defer = self._defer()
            if not did_defer:
                self.fail(f"No concrete models found for {names}")
            return None

        return self._make_union(is_type, tuple(concrete))

    def _make_union(
        self, is_type: bool, instances: Sequence[Instance | PlaceholderType]
    ) -> UnionType | Instance | TypeType | PlaceholderType:
        """
        Given a sequence of instances, make them all TypeType if is_type and then
        return either the one type if the list is of 1, or the list wrapped in a UnionType
        """
        items: Sequence[UnionType | TypeType | Instance | PlaceholderType]

        if is_type:
            items = [TypeType(item) for item in instances]
        else:
            items = instances

        if len(items) == 1:
            return items[0]
        else:
            return UnionType(tuple(items))

    def _instances_from_aliases(
        self, get_aliases: protocols.AliasGetter, *models: str
    ) -> Iterator[Instance | PlaceholderType]:
        for model, alias in get_aliases(*models).items():
            if alias is None:
                self.fail(f"Failed to find concrete alias instance for '{model}'")
                continue

            try:
                yield from self.lookup_alias(alias)
            except FailedLookup as error:
                self.fail(
                    f"Failed to create concrete alias instance for '{model}' ({error}) (this is likely a bug in extended_mypy_django_plugin)"
                )

    def resolve(
        self, annotation: protocols.KnownAnnotations, model_type: ProperType
    ) -> Instance | TypeType | UnionType | PlaceholderType | None:
        if annotation is protocols.KnownAnnotations.CONCRETE:
            return self._concrete_for(model_type, self.get_concrete_aliases)

        elif annotation is protocols.KnownAnnotations.DEFAULT_QUERYSET:
            return self._concrete_for(model_type, self.get_queryset_aliases)

        else:
            assert_never(annotation)


make_resolver = AnnotationResolver.create

if TYPE_CHECKING:
    _R: protocols.Resolver = cast(AnnotationResolver, None)



from django.db import models
from typing_extensions import Self

from extended_mypy_django_plugin import Concrete


class Parent(models.Model):
    one = models.CharField(max_length=50)

    @classmethod
    def new(cls, one: str) -> Self:
        concrete = Concrete.cast_as_concrete(cls)
        created = concrete.objects.create(one=one)
        assert isinstance(created, cls)
        return created

    class Meta:
        abstract = True


class Child1(Parent):
    two = models.CharField(max_length=60)


class Child2QuerySet(models.QuerySet["Child2"]):
    pass


Child2Manager = models.Manager.from_queryset(Child2QuerySet)


class Child2(Parent):
    two = models.CharField(max_length=60)
    four = models.CharField(max_length=1)

    three = models.CharField(max_length=70)

    objects = Child2Manager()


class Parent2(Parent):
    three = models.CharField(max_length=50)

    class Meta:
        abstract = True


class Child3(Parent2):
    two = models.CharField(max_length=60)

    three = models.CharField(max_length=70)


class Child4QuerySet(models.QuerySet["Child2"]):
    pass


Child4Manager = models.Manager.from_queryset(Child4QuerySet)


class Child4(Parent2):
    two = models.CharField(max_length=60)

    three = models.CharField(max_length=70)

    objects = Child4Manager()



import zlib
from typing import TYPE_CHECKING

from . import protocols


def adler32_hash(*parts: bytes) -> str:
    return str(zlib.adler32(b"\n".join(parts)))


if TYPE_CHECKING:
    _a32h: protocols.Hasher = adler32_hash



def interface____differentiated__12() -> None:
    return None

mod = "django.contrib.sessions.base_session"
summary = "__virtual__.mod_113708644::django.contrib.sessions.base_session::installed_apps=__installed_apps_hash__::significant=1879365353"

import django.contrib.sessions.base_session
import django.contrib.sessions.models
import django.db.models
ConcreteQuerySet__AbstractBaseSession = django.db.models.QuerySet[django.contrib.sessions.models.Session]
Concrete__AbstractBaseSession = django.contrib.sessions.models.Session



def interface____differentiated__7() -> None:
    return None

mod = "djangoexample.relations1.models"
summary = "__virtual__.mod_3327724610::djangoexample.relations1.models::installed_apps=__installed_apps_hash__::significant=448140218"

import django.db.models
import djangoexample.relations1.models
ConcreteQuerySet__Abstract = djangoexample.relations1.models.Child1QuerySet | django.db.models.QuerySet[djangoexample.relations1.models.Child2]
ConcreteQuerySet__Child1 = djangoexample.relations1.models.Child1QuerySet
ConcreteQuerySet__Child2 = django.db.models.QuerySet[djangoexample.relations1.models.Child2]
ConcreteQuerySet__Concrete1 = djangoexample.relations1.models.Concrete1QuerySet
ConcreteQuerySet__Concrete2 = django.db.models.QuerySet[djangoexample.relations1.models.Concrete2]
Concrete__Abstract = djangoexample.relations1.models.Child1 | djangoexample.relations1.models.Child2
Concrete__Child1 = djangoexample.relations1.models.Child1
Concrete__Child2 = djangoexample.relations1.models.Child2
Concrete__Concrete1 = djangoexample.relations1.models.Concrete1
Concrete__Concrete2 = djangoexample.relations1.models.Concrete2



import dataclasses

from extended_mypy_django_plugin.django_analysis import (
    ImportPath,
    Project,
    protocols,
    virtual_dependencies,
)


@dataclasses.dataclass
class Namer:
    namespace: protocols.ImportPath = dataclasses.field(
        default_factory=lambda: ImportPath("__virtual__")
    )

    def __call__(self, module: protocols.ImportPath, /) -> protocols.ImportPath:
        return ImportPath(f"{self.namespace}.mod_{module.replace('.','_')}")


class TestVirtualDependency:
    def test_making_virtual_dependency(
        self, discovered_django_example: protocols.Discovered[Project]
    ) -> None:
        import djangoexample.exampleapp.models

        module = discovered_django_example.installed_models_modules[
            ImportPath.from_module(djangoexample.exampleapp.models)
        ]

        related_models: set[protocols.ImportPath] = {
            ImportPath("djangoexample.exampleapp.models.Child3"),
            ImportPath("djangoexample.exampleapp.models.Parent"),
            ImportPath("djangoexample.exampleapp.models.Child2"),
            ImportPath("djangoexample.exampleapp.models.Child1"),
            ImportPath("djangoexample.exampleapp.models.Parent2"),
            ImportPath("djangoexample.exampleapp.models.Child4"),
        }

        significant_info = [
            "module:djangoexample.exampleapp.models",
            "module:djangoexample.exampleapp.models>concrete:djangoexample.exampleapp.models.Parent=djangoexample.exampleapp.models.Child1,djangoexample.exampleapp.models.Child2,djangoexample.exampleapp.models.Child3,djangoexample.exampleapp.models.Child4,djangoexample.exampleapp2.models.ChildOther,djangoexample.exampleapp2.models.ChildOther2",
            "module:djangoexample.exampleapp.models>concrete:djangoexample.exampleapp.models.Parent2=djangoexample.exampleapp.models.Child3,djangoexample.exampleapp.models.Child4",
            "module:djangoexample.exampleapp.models>concrete:djangoexample.exampleapp.models.Child1=djangoexample.exampleapp.models.Child1",
            "module:djangoexample.exampleapp.models>concrete:djangoexample.exampleapp.models.Child2=djangoexample.exampleapp.models.Child2",
            "module:djangoexample.exampleapp.models>concrete:djangoexample.exampleapp.models.Child3=djangoexample.exampleapp.models.Child3",
            "module:djangoexample.exampleapp.models>concrete:djangoexample.exampleapp.models.Child4=djangoexample.exampleapp.models.Child4",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Parent>is_abstract:True",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Parent>field:one",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Parent>field:one>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Parent2>is_abstract:True",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Parent2>mro_0:djangoexample.exampleapp.models.Parent",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Parent2>field:one",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Parent2>field:one>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Parent2>field:three",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Parent2>field:three>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child1>is_abstract:False",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child1>mro_0:djangoexample.exampleapp.models.Parent",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child1>field:id",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child1>field:id>field_type:django.db.models.fields.BigAutoField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child1>field:one",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child1>field:one>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child1>field:two",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child1>field:two>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>is_abstract:False",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>custom_queryset:djangoexample.exampleapp.models.Child2QuerySet",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>mro_0:djangoexample.exampleapp.models.Parent",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:id",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:id>field_type:django.db.models.fields.BigAutoField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:one",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:one>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:two",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:two>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:four",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:four>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:three",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child2>field:three>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>is_abstract:False",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>mro_0:djangoexample.exampleapp.models.Parent2",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>mro_1:djangoexample.exampleapp.models.Parent",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>field:id",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>field:id>field_type:django.db.models.fields.BigAutoField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>field:one",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>field:one>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>field:two",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>field:two>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>field:three",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child3>field:three>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>is_abstract:False",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>custom_queryset:djangoexample.exampleapp.models.Child4QuerySet",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>mro_0:djangoexample.exampleapp.models.Parent2",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>mro_1:djangoexample.exampleapp.models.Parent",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>field:id",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>field:id>field_type:django.db.models.fields.BigAutoField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>field:one",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>field:one>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>field:two",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>field:two>field_type:django.db.models.fields.CharField",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>field:three",
            "module:djangoexample.exampleapp.models>model:djangoexample.exampleapp.models.Child4>field:three>field_type:django.db.models.fields.CharField",
        ]

        virtual_dependency = virtual_dependencies.VirtualDependency.create(
            discovered_project=discovered_django_example,
            module=module,
            virtual_dependency_namer=Namer(),
        )

        all_models = discovered_django_example.all_models

        assert virtual_dependency == virtual_dependencies.VirtualDependency(
            module=module,
            summary=virtual_dependencies.VirtualDependencySummary(
                virtual_namespace=ImportPath("__virtual__"),
                virtual_import_path=ImportPath("__virtual__.mod_djangoexample_exampleapp_models"),
                module_import_path=module.import_path,
                significant_info=significant_info,
            ),
            all_related_models=sorted(related_models),
            concrete_models={
                ImportPath("djangoexample.exampleapp.models.Parent"): [
                    all_models[ImportPath("djangoexample.exampleapp.models.Child1")],
                    all_models[ImportPath("djangoexample.exampleapp.models.Child2")],
                    all_models[ImportPath("djangoexample.exampleapp.models.Child3")],
                    all_models[ImportPath("djangoexample.exampleapp.models.Child4")],
                    all_models[ImportPath("djangoexample.exampleapp2.models.ChildOther")],
                    all_models[ImportPath("djangoexample.exampleapp2.models.ChildOther2")],
                ],
                ImportPath("djangoexample.exampleapp.models.Parent2"): [
                    all_models[ImportPath("djangoexample.exampleapp.models.Child3")],
                    all_models[ImportPath("djangoexample.exampleapp.models.Child4")],
                ],
                ImportPath("djangoexample.exampleapp.models.Child1"): [
                    all_models[ImportPath("djangoexample.exampleapp.models.Child1")]
                ],
                ImportPath("djangoexample.exampleapp.models.Child2"): [
                    all_models[ImportPath("djangoexample.exampleapp.models.Child2")]
                ],
                ImportPath("djangoexample.exampleapp.models.Child3"): [
                    all_models[ImportPath("djangoexample.exampleapp.models.Child3")]
                ],
                ImportPath("djangoexample.exampleapp.models.Child4"): [
                    all_models[ImportPath("djangoexample.exampleapp.models.Child4")]
                ],
            },
        )

    def test_significant_info_takes_into_account_field_related_model(
        self, discovered_django_example: protocols.Discovered[Project]
    ) -> None:
        import djangoexample.relations1.models

        module = discovered_django_example.installed_models_modules[
            ImportPath.from_module(djangoexample.relations1.models)
        ]

        virtual_dependency = virtual_dependencies.VirtualDependency.create(
            discovered_project=discovered_django_example,
            module=module,
            virtual_dependency_namer=Namer(),
        )

        assert virtual_dependency.summary.significant_info == [
            "module:djangoexample.relations1.models",
            "module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Abstract=djangoexample.relations1.models.Child1,djangoexample.relations1.models.Child2",
            "module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Child1=djangoexample.relations1.models.Child1",
            "module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Child2=djangoexample.relations1.models.Child2",
            "module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Concrete1=djangoexample.relations1.models.Concrete1",
            "module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Concrete2=djangoexample.relations1.models.Concrete2",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Abstract>is_abstract:True",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>is_abstract:False",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>custom_queryset:djangoexample.relations1.models.Child1QuerySet",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>mro_0:djangoexample.relations1.models.Abstract",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:Concrete2_children+",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:Concrete2_children+>field_type:django.db.models.fields.reverse_related.ManyToOneRel",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:children",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:children>field_type:django.db.models.fields.reverse_related.ManyToManyRel",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:children>related_model:djangoexample.relations1.models.Concrete2",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:id",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:id>field_type:django.db.models.fields.BigAutoField",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child2>is_abstract:False",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child2>mro_0:djangoexample.relations1.models.Abstract",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child2>field:id",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child2>field:id>field_type:django.db.models.fields.BigAutoField",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>is_abstract:False",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>custom_queryset:djangoexample.relations1.models.Concrete1QuerySet",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:c2s",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:c2s>field_type:django.db.models.fields.reverse_related.ManyToOneRel",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:c2s>related_model:djangoexample.relations1.models.Concrete2",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:thing",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:thing>field_type:django.db.models.fields.reverse_related.OneToOneRel",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:thing>related_model:djangoexample.relations2.models.Thing",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:id",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:id>field_type:django.db.models.fields.BigAutoField",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>is_abstract:False",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:Concrete2_children+",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:Concrete2_children+>field_type:django.db.models.fields.reverse_related.ManyToOneRel",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:id",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:id>field_type:django.db.models.fields.BigAutoField",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:concrete1",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:concrete1>field_type:django.db.models.fields.related.ForeignKey",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:concrete1>related_model:djangoexample.relations1.models.Concrete1",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:children",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:children>field_type:django.db.models.fields.related.ManyToManyField",
            "module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:children>related_model:djangoexample.relations1.models.Child1",
        ]



import pytest

from extended_mypy_django_plugin.django_analysis import ImportPath, discovery, protocols
from extended_mypy_django_plugin.django_analysis.discovery.import_path import ImportPathHelper


class TestImportPathHelper:
    def test_an_instance_is_provided(self) -> None:
        assert isinstance(ImportPath, ImportPathHelper)

    def test_can_return_path_to_class(self) -> None:
        assert (
            ImportPath.from_cls(TestImportPathHelper)
            == "tests.django_analysis.discovery.test_import_path.TestImportPathHelper"
        )
        assert (
            ImportPath.from_cls(ImportPathHelper)
            == "extended_mypy_django_plugin.django_analysis.discovery.import_path.ImportPathHelper"
        )

    def test_can_return_module_import_path_from_class(self) -> None:
        assert (
            ImportPath.cls_module(TestImportPathHelper)
            == "tests.django_analysis.discovery.test_import_path"
        )
        assert (
            ImportPath.cls_module(ImportPathHelper)
            == "extended_mypy_django_plugin.django_analysis.discovery.import_path"
        )

    def test_can_split_an_import_path(self) -> None:
        assert ImportPath.split(protocols.ImportPath("somewhere.nice")) == (
            protocols.ImportPath("somewhere"),
            protocols.ImportPath("nice"),
        )
        assert ImportPath.split(protocols.ImportPath("somewhere.else.that.is.good")) == (
            protocols.ImportPath("somewhere.else.that.is"),
            protocols.ImportPath("good"),
        )

    def test_complains_if_splitting_non_namespaced_import(self) -> None:
        with pytest.raises(discovery.InvalidImportPath):
            ImportPath.split(protocols.ImportPath("not_namespaced"))

    def test_turns_string_into_import_path(self) -> None:
        assert ImportPath("hello.there") == protocols.ImportPath("hello.there")
        assert ImportPath("place") == protocols.ImportPath("place")

    @pytest.mark.parametrize(
        "invalid",
        (
            pytest.param(".wat", id="starts_with_dot"),
            pytest.param("stuff..things", id="double_dot"),
            pytest.param("3stuff", id="starts_with_number"),
            pytest.param("things-stuff", id="dashes_not_valid_identifier"),
            pytest.param("", id="empty_string"),
            pytest.param("%%%", id="another_invalid_identifier"),
        ),
    )
    def test_complains_if_string_is_not_valid_import_path(self, invalid: str) -> None:
        with pytest.raises(discovery.InvalidImportPath):
            ImportPath(invalid)



import dataclasses
import functools
import os
import pathlib
from typing import TYPE_CHECKING, Generic, cast

from .. import project, protocols
from . import dependency, report


@dataclasses.dataclass(frozen=True, kw_only=True)
class VirtualDependencyGenerator(Generic[protocols.T_Project, protocols.T_VirtualDependency]):
    virtual_dependency_maker: protocols.VirtualDependencyMaker[
        protocols.T_Project, protocols.T_VirtualDependency
    ]

    def __call__(
        self, *, discovered_project: protocols.Discovered[protocols.T_Project]
    ) -> protocols.VirtualDependencyMap[protocols.T_VirtualDependency]:
        return {
            import_path: self.virtual_dependency_maker(
                discovered_project=discovered_project, module=module
            )
            for import_path, module in discovered_project.installed_models_modules.items()
        }


@dataclasses.dataclass(frozen=True, kw_only=True)
class VirtualDependencyInstaller(Generic[protocols.T_VirtualDependency, protocols.T_Report]):
    project_version: str
    virtual_dependencies: protocols.VirtualDependencyMap[protocols.T_VirtualDependency]
    virtual_dependency_namer: protocols.VirtualDependencyNamer

    def __call__(
        self,
        *,
        scratch_root: pathlib.Path,
        destination: pathlib.Path,
        virtual_namespace: protocols.ImportPath,
        report_factory: protocols.ReportFactory[protocols.T_VirtualDependency, protocols.T_Report],
    ) -> protocols.CombinedReport[protocols.T_Report]:
        # Determine what would be written to represent the virtual dependencies
        # And gather each report so we can later combine them into the final report
        reports: list[protocols.T_Report] = []
        written_dependencies: list[protocols.RenderedVirtualDependency[protocols.T_Report]] = []
        for rendered in report_factory.deploy_scribes(self.virtual_dependencies):
            report_factory.report_installer.write_report(
                virtual_import_path=rendered.virtual_import_path,
                summary_hash=rendered.summary_hash,
                content=rendered.content,
                scratch_root=scratch_root,
            )
            written_dependencies.append(rendered)
            reports.append(rendered.report)

        # Install our on disk representation into the destination
        report_factory.report_installer.install_reports(
            scratch_root=scratch_root,
            destination=destination,
            virtual_namespace=virtual_namespace,
        )

        # Figure out a string representing the state of everything
        version = report_factory.determine_version(
            destination=destination,
            virtual_namespace=virtual_namespace,
            project_version=self.project_version,
            written_dependencies=written_dependencies,
        )

        # Create our final report
        return report_factory.report_combiner_maker(reports=reports).combine(
            version=version,
            write_empty_virtual_dep=functools.partial(
                self.write_empty_virtual_dep,
                destination=destination,
                report_factory=report_factory,
            ),
        )

    def write_empty_virtual_dep(
        self,
        *,
        module_import_path: protocols.ImportPath,
        destination: pathlib.Path,
        report_factory: protocols.ReportFactory[protocols.T_VirtualDependency, protocols.T_Report],
    ) -> protocols.ImportPath | None:
        """
        This method gets used at mypy time when we come across modules that we believe contains Django
        models that aren't included in settings.INSTALLED_APPS

        This is so when they are put into settings.INSTALLED_APPS, they don't also need to be modified
        to have a virtual dependency we can use to tell mypy to re-analyze the file on relevant changes
        in other parts of the codebase.
        """
        virtual_import_path = self.virtual_dependency_namer(module_import_path)
        location = destination / f'{virtual_import_path.replace(".", os.sep)}.py'
        if location.exists():
            return None

        content = report_factory.make_empty_virtual_dependency_content(
            module_import_path=module_import_path
        )
        if report_factory.report_installer.write_report(
            scratch_root=destination,
            virtual_import_path=virtual_import_path,
            content=content,
            summary_hash=False,
        ):
            return virtual_import_path
        else:
            return None


if TYPE_CHECKING:
    C_VirtualDependencyGenerator = VirtualDependencyGenerator[
        project.C_Project, dependency.C_VirtualDependency
    ]
    C_VirtualDependencyInstaller = VirtualDependencyInstaller[
        dependency.C_VirtualDependency, report.C_Report
    ]

    _VDN: protocols.P_VirtualDependencyGenerator = cast(
        VirtualDependencyGenerator[protocols.P_Project, protocols.P_VirtualDependency],
        None,
    )
    _GVD: protocols.P_VirtualDependencyInstaller = cast(
        VirtualDependencyInstaller[protocols.P_VirtualDependency, protocols.P_Report], None
    )

    _CVDN: protocols.VirtualDependencyGenerator[
        project.C_Project, dependency.C_VirtualDependency
    ] = cast(C_VirtualDependencyGenerator, None)
    _CGVD: protocols.VirtualDependencyInstaller[
        dependency.C_VirtualDependency, report.C_Report
    ] = cast(C_VirtualDependencyInstaller, None)



def interface____differentiated__4() -> None:
    return None

mod = "django.contrib.sessions.models"
summary = "__virtual__.mod_3074165738::django.contrib.sessions.models::installed_apps=__installed_apps_hash__::significant=4231118930"

import django.contrib.sessions.models
import django.db.models
ConcreteQuerySet__Session = django.db.models.QuerySet[django.contrib.sessions.models.Session]
Concrete__Session = django.contrib.sessions.models.Session



from __future__ import annotations

from typing import Generic, TypeVar

from django.db import models

T_Parent = TypeVar("T_Parent")
T_Obj = TypeVar("T_Obj")


class Concrete(Generic[T_Parent]):
    """
    The ``Concrete`` annotation exists as a class with functionality for both
    runtime and static type checking time.

    At static type checking time (specifically with ``mypy``) it is used to create
    a type that represents the Union of all the concrete children of some
    abstract model.

    .. automethod:: cast_as_concrete
    """

    @classmethod
    def cast_as_concrete(cls, obj: T_Obj) -> T_Obj:
        """
        This can be used to change the type of an abstract django model to be only
        a concrete decedent.

        At runtime this will raise an exception if the object is an abstract model or class.

        At static type checking time this will change the type of the variable being assigned to::

            from typing import Self, cast
            from extended_mypy_django_plugin import Concrete

            class MyAbstractModel(Model):
                class Meta:
                    abstract = True

                @classmethod
                def new(cls) -> Self:
                    cls = Concrete.cast_as_concrete(cls)
                    reveal_type(cls) # type[Concrete1] | type[Concrete2] | type[Concrete3] | ...
                    ...

                def get_self(self) -> Self:
                    self = Concrete.cast_as_concrete(self)
                    reveal_type(self) # Concrete1 | Concrete2 | Concrete3 | ...
                    ...

        This can also be used outside of a model method::

            model: type[MyAbstractModel] = Concrete1
            narrowed = Concrete.cast_as_concrete(model)
            reveal_type(narrowed) # Concrete1 | Concrete2 | Concrete3 | ...
        """
        if isinstance(obj, type):
            if not issubclass(obj, models.Model) or (
                (Meta := getattr(obj, "Meta", None)) and getattr(Meta, "abstract", False)
            ):
                raise RuntimeError("Expected a concrete subclass")

        elif not isinstance(obj, models.Model) or obj._meta.abstract:
            raise RuntimeError("Expected a concrete instance")

        return obj


class DefaultQuerySet(Generic[T_Parent]):
    """
    This is used to annotate a model such that the mypy plugin may turn this into
    a union of all the default querysets for all the concrete children of the
    specified abstract model class, or of that model when it is a concrete model
    """



from django.db import models


class Abstract(models.Model):
    class Meta:
        abstract = True


class Child1QuerySet(models.QuerySet["Child1"]):
    pass


Child1Manager = models.Manager.from_queryset(Child1QuerySet)


class Child1(Abstract):
    objects = Child1Manager()


class Child2(Abstract):
    pass


class Concrete1QuerySet(models.QuerySet["Concrete1"]):
    pass


Concrete1Manager = models.Manager.from_queryset(Concrete1QuerySet)


class Concrete1(models.Model):
    objects = Concrete1Manager()


class Concrete2(models.Model):
    concrete1 = models.ForeignKey(
        "relations1.Concrete1", related_name="c2s", on_delete=models.CASCADE
    )
    children = models.ManyToManyField(Child1, related_name="children")



def interface____differentiated__5() -> None:
    return None

mod = "djangoexample.exampleapp.models"
summary = "__virtual__.mod_3347844205::djangoexample.exampleapp.models::installed_apps=__installed_apps_hash__::significant=2358377148"

import django.db.models
import djangoexample.exampleapp.models
import djangoexample.exampleapp2.models
ConcreteQuerySet__Child1 = django.db.models.QuerySet[djangoexample.exampleapp.models.Child1]
ConcreteQuerySet__Child2 = djangoexample.exampleapp.models.Child2QuerySet
ConcreteQuerySet__Child3 = django.db.models.QuerySet[djangoexample.exampleapp.models.Child3]
ConcreteQuerySet__Child4 = djangoexample.exampleapp.models.Child4QuerySet
ConcreteQuerySet__Parent = django.db.models.QuerySet[djangoexample.exampleapp.models.Child1] | djangoexample.exampleapp.models.Child2QuerySet | django.db.models.QuerySet[djangoexample.exampleapp.models.Child3] | djangoexample.exampleapp.models.Child4QuerySet | django.db.models.QuerySet[djangoexample.exampleapp2.models.ChildOther] | django.db.models.QuerySet[djangoexample.exampleapp2.models.ChildOther2]
ConcreteQuerySet__Parent2 = django.db.models.QuerySet[djangoexample.exampleapp.models.Child3] | djangoexample.exampleapp.models.Child4QuerySet
Concrete__Child1 = djangoexample.exampleapp.models.Child1
Concrete__Child2 = djangoexample.exampleapp.models.Child2
Concrete__Child3 = djangoexample.exampleapp.models.Child3
Concrete__Child4 = djangoexample.exampleapp.models.Child4
Concrete__Parent = djangoexample.exampleapp.models.Child1 | djangoexample.exampleapp.models.Child2 | djangoexample.exampleapp.models.Child3 | djangoexample.exampleapp.models.Child4 | djangoexample.exampleapp2.models.ChildOther | djangoexample.exampleapp2.models.ChildOther2
Concrete__Parent2 = djangoexample.exampleapp.models.Child3 | djangoexample.exampleapp.models.Child4
